

from numpy import sum as sum_, array, floor
from pandas import Series,concat,DataFrame, to_datetime
from datetime import datetime,timedelta
from klp_commons.controllers.controller_mongodb import ControllerMongoDB
from klp_commons.controllers.controller_postgresdb import ControllerPostgresDB

class Summary:
    """
    """

    # Variables de clase

    def __init__(self,from_date_str = None, num_days= 30):
        self.CMongo = ControllerMongoDB()
        self.CMongo.get_con_db('klopp')
        
        self.CPDB = ControllerPostgresDB()

        self.category_names = self.CPDB.get_categorys_names()
        self.subcategory_names = self.CPDB.get_subcategorys_names()
        
        self.dict_names = {k:v for k,v in self.category_names}

        self.rename_cols_ = [
                    'user_id', 
                    'id',
                    'description',
                    'clean_description',
                    'original_created_at',
                    'validated',
                    'bank',
                    'subcategory_precategorized',
                    'subcagegory_custom',
                    'category_precategorized',
                    'category_custom',
                    'nickname_suggested',
                    'nickname_custom',
                    'change_nickname',
                    'change_category',
                    'change_subcategory',
                    'updated_num',
                    'classification_date',
                    'classification_time',
                    'created_at_date',
                    'created_at_time',
                    'updated_at_date',
                    'updated_at_time'
                            ]
        self.rename_cols = [
                    'user_id', 
                    'id',
                    'description',
                    'clean_description',
                    'classified_at',
                    'original_created_at',
                    'created_at',
                    'updated_at',
                    'validated',
                    'bank',
                    'subcategory_precategorized',
                    'subcagegory_custom',
                    'category_precategorized',
                    'category_custom',
                    'nickname_suggested',
                    'nickname_custom',
                    'change_nickname',
                    'change_category',
                    'change_subcategory',
                    'updated_num'
                    ]
        self.num_days = num_days
        self.last_period = None
        self.from_date = from_date_str
        self.to_date = None
        self.set_date_range()
        
    def set_date_range(self):
        self.from_date = datetime.strptime(self.from_date , '%Y-%m-%d')
        self.to_date = self.from_date + timedelta(days=self.num_days)
        self.last_period = self.from_date - timedelta(days=self.num_days)

    def agg_master(self):
        self.query_summary_category = [
            # { "$unwind": "$history-track" },
            {
            "$match": {
                "$or":[
                    {"created_at": {'$gte':self.from_date,'$lt': self.to_date}},
                    {"history-track.updated_at": {'$gte':self.from_date,'$lt': self.to_date}}
                        ]
                    }
            },
        {'$group': 
                 {'_id': {
                        'user_id': "$user_uid",
                        'id':'$_id',
                        'original_created_at': "$original_created_at",
                        'classified_at': "$history-track.classified_at",
                        'created_at': "$created_at" ,
                        'transaction_type': "$type" ,
                        'updated_at': "$history-track.updated_at",
                        'description':'$source_description',
                        'clean_description':'$history-track.clean_description',
                        'category_id': "$history-track.category_uid",
                        'nickname': '$history-track.nickname',
                        'type': "$history-track.type",
                        'bank': "$account_institution_name",
                        'change_nickname': '$history-track.change_nickname',
                        'change_subcategory': '$history-track.change_subcat',
                        'change_category': '$history-track.change_cat',
                        'validated': '$history-track.validated',
                       },
                'count': {"$sum": 1}
                }

             
            }
            
        ]

        
        
        self.agg_result = self.CMongo.con_db.fact.aggregate(self.query_summary_category) 

        df_list= list()
        
        for i in self.agg_result: 
            df_tmp = DataFrame(i)
            count = df_tmp['count'] 
            del df_tmp['count']
            df_tmp.loc['count', :] = [count.values]
            df_list.append(DataFrame(df_tmp).transpose()) 

        self.df_agg_master = concat(df_list) 
        self.df_agg_master.reset_index(inplace = True, drop = True)
        self.df_agg_master['count'] = self.df_agg_master['count'].apply(Series)[1]
        
        del self.df_agg_master['count']
        
        self.df_agg_master['updated_at'].fillna(False,inplace = True)
        #self.df_agg_master['validated'] = self.df_agg_master['validated'].fillna(False)
    def summary_master(self):
        
        self.agg_master()
        df_summ = self.df_agg_master 
        df_summ.reset_index(inplace = True)
        df_summ['category_name'] = self.map_names(df_summ.category_id.values.tolist())

        # error , se elimino subcategoryname de el flujo de trabajo de categorization
        # ya no se almacena; es necesario hacer un cruce con el subcategoyr uid
        tmp1 = df_summ['subcategory_name'].apply(Series).add_prefix('SUBCATEGORY_NAME_LEVEL_')[['SUBCATEGORY_NAME_LEVEL_0','SUBCATEGORY_NAME_LEVEL_1']]
        
        tmp2 = df_summ['category_name'].apply(Series).add_prefix('CATEGORY_NAME_LEVEL_')[['CATEGORY_NAME_LEVEL_0','CATEGORY_NAME_LEVEL_1']]
        
        tmp3 =df_summ['nickname'].apply(Series).add_prefix('NICKNAME_LEVEL_')[['NICKNAME_LEVEL_0','NICKNAME_LEVEL_1']]

        tmp4 = df_summ['change_nickname'].apply(Series,dtype=object).add_prefix('change_nickname_LEVEL_')[['change_nickname_LEVEL_0']]
        tmp5 = df_summ['change_subcategory'].apply(Series,dtype=object).add_prefix('change_subcategory_LEVEL_')[['change_subcategory_LEVEL_0']]
        tmp6 =df_summ['change_category'].apply(Series,dtype=object).add_prefix('change_category_LEVEL_')[['change_category_LEVEL_0']]
        
        
        df_summ['validated'] = df_summ['validated'].apply(Series,dtype=object).add_prefix('validated_LEVEL_')[['validated_LEVEL_0']]
        
        df_summ['updated_at'] = df_summ['updated_at'].apply(Series,dtype=object).add_prefix('updated_at_LEVEL_')[['updated_at_LEVEL_1']]
        
        df_tmp = df_summ['clean_description'].apply(Series,dtype=object)
        
        if df_tmp.empty :
            df_summ['clean_description'] = df_summ['description'].apply(Series,dtype=object)
        else : 
            df_summ['clean_description'] = df_summ['clean_description'].apply(Series,dtype=object)
            
        df_tmp = df_summ['classified_at'].apply(Series,dtype=object)
        
        if df_tmp.empty : 
            df_summ['classified_at'] =  df_summ['updated_at']
        else :
            df_summ['classified_at'] = df_summ['classified_at'].apply(Series,dtype=object)

        df_expand = df_summ['type'].apply(Series)
        df_expand = df_expand.apply(Series.value_counts, axis=1)

        if 'report-expense-updated' in df_expand.columns :
            df_expand = df_expand[['transaction-categorized','report-expense-created','report-expense-updated']]
            df_expand.columns = ['pre-categorized','validate_','updated']
            df_expand = df_expand['updated']
        else :
            
            df_expand['updated'] = 0
            del df_expand['transaction-categorized']
            del df_expand['report-expense-created']
            
            
            
        del  df_summ['type']
        
        cols = ['user_id',
                'id',
                'description',
                'clean_description',
                'classified_at',
                'original_created_at',
                'created_at',
                'updated_at',
                'validated',
                'bank',
               'transaction_type']
        
        
        frames = [df_summ[cols],tmp1,tmp2,tmp3,tmp4,tmp6,tmp5,df_expand]

        df_final = concat(frames, axis=1)
        df_final.fillna(False,inplace=True)
        
        df_final['bank'] = df_final.bank.str.split(pat = '_', expand = True)[0]
        
        self.df_master = df_final

    def trasform_datetime(self):
        
        self.df_master['classification_date'] = to_datetime(s.df_master.classified_at).dt.date
        self.df_master['classification_time'] = to_datetime(s.df_master.classified_at).dt.time
        
        self.df_master['created_at_date'] = to_datetime(s.df_master.classified_at).dt.date
        self.df_master['created_at_time'] = to_datetime(s.df_master.classified_at).dt.time
        
        
        self.df_master['updated_at_date'] = to_datetime(s.df_master.classified_at).dt.date
        self.df_master['updated_at_time'] = to_datetime(s.df_master.classified_at).dt.time     
        
    def remove_conflicts(self):
        ids_list = self.df_master[self.df_master.created_at<self.from_date].id.values.tolist()
        
        self.CMongo.con_db.AggMaster.cube.delete_many({"ID": { "$in": ids_list}})
        self.CMongo.con_db.FacMaster.cube.delete_many({"ID": { "$in": ids_list}}) 
        
        #del self.df_master['classified_at']
        #del self.df_master['created_at']
        #del self.df_master['updated_at']
        self.df_master.columns = self.rename_cols
   
    def add_new_data(self):
        self.df_agg_master['updated_at'].fillna(False,inplace = True)
        self.CMongo.con_db.AggMaster.cube.insert_many(self.df_agg_master.to_dict(orient='records'))
        self.CMongo.con_db.FacMaster.cube.insert_many(self.df_master.to_dict(orient='records'))

    def map_names(self,list_nested_uids):
        list_final = list ()
        for l in list_nested_uids:
            list_nested =list()
            for item in l:
                item = self.dict_names.get(item,item)
                list_nested.append(item)
            
            list_final.append(list_nested)
        return DataFrame(array(list_final,dtype=object))
      
    def summary_category_lv1 (self):
        columns_cat_level_1 = ['Category',
                               'Users (With this category used)',
                               'Classified transactions',
                               'Transaction per review']
        
        grouped = self.df_master .groupby(['category_precategorized'])
        
        summary_gruped = grouped.agg({'user_id': Series.nunique,
                             'id': Series.nunique,
                             'validated': Series.sum})
        
        summary_gruped.reset_index(inplace = True)
        summary_gruped['validated'] = summary_gruped['id'] - summary_gruped['validated'] 
        summary_gruped.columns = columns_cat_level_1
        
        return summary_gruped
        
    def summary_nickname_lv1(self):
        
        columns_nickname_level_1 = [
                            'Precategorized',
                            'Nicknames suggesteds',
                            'Number of changes nicknames',
                            'Percentage of changes nicknames',
                            'Nickname accuracy percentage']  
        
        #self.df_master['nickname_detected'] = self.df_master['nickname_suggested'] != ''
        
        grouped = self.df_master .groupby(['category_precategorized'])
        summary_gruped = grouped.agg({
                'nickname_suggested': Series.count,
                 'change_nickname': Series.sum,

        })
        
        summary_gruped.reset_index(inplace = True)
        
        summary_gruped['Percentage of changes'] = (summary_gruped['change_nickname'] *100/ 
              summary_gruped['nickname_suggested']).apply(floor)
        tmp  =summary_gruped['nickname_suggested']-summary_gruped['change_nickname']
        
        summary_gruped['Nickname accuracy percentage'] = (tmp/summary_gruped['nickname_suggested']*100).apply(floor)
        

        summary_gruped.columns = columns_nickname_level_1
        
        del summary_gruped['Precategorized']
        return summary_gruped
    
    def summary_subcategory_lv1(self):
        columns_subcat_level_1 = ['category_precategorized',
        'Subcategory suggested',
        'Number of changes subcategory',
        'Percentage of changes subcategory',
        'Subcategory accuracy percent'
                          ]

    
        grouped = self.df_master .groupby(['category_precategorized'])
        
        summary_gruped = grouped.agg({'subcategory_precategorized': Series.count,
             'change_subcategory': Series.sum})
        summary_gruped.reset_index(inplace = True)
        
        summary_gruped['Percentage of changes'] = (summary_gruped['change_subcategory'] *100/ 
          summary_gruped['subcategory_precategorized']).apply(floor)
        tmp  =summary_gruped['subcategory_precategorized']-summary_gruped['change_subcategory']
        summary_gruped['Subcategory accuracy percentage'] = (tmp/summary_gruped['subcategory_precategorized']*100).apply(floor)
        summary_gruped.columns = columns_subcat_level_1
        
        del summary_gruped['category_precategorized']
        return summary_gruped
    
    def summary_category_lv2(self):

        columns_cat_level_2 = [
                'Category',
                'Subcategory',
                'Classified transactions',
                'Users (With this subcategory used)',
                'Unclassified transactions'
                ]
        
        self.df_master['classified'] = self.df_master['subcategory_precategorized'] != 'Without Classification'
        self.df_master['unclassified'] = self.df_master['subcategory_precategorized'] == 'Without Classification'

        grouped = self.df_master .groupby(['category_precategorized','subcategory_precategorized'])
        summary_gruped = grouped.agg({
             'classified': Series.sum,
             'user_id': Series.nunique,
            'unclassified':Series.sum})

        summary_gruped.reset_index(inplace = True)
        summary_gruped.columns = columns_cat_level_2
        del summary_gruped['Category']
        
        return summary_gruped

    def summary_nickname_lv2(self):

        columns_nickname_level_2 = [
                    'Category',
                    'Subcategory',
                    'Nicknames suggesteds',
                    'Number of changes nicknames',
                    'Percentage of changes nicknames',
                    'Nickname accuracy percent'
                          ]
        
        sub_cols = [
                'Nicknames suggesteds',
                'Number of changes nicknames',
                'Percentage of changes nicknames',
                'Nickname accuracy percent'
                ]
        
        #self.df_master['nickname_detected'] = self.df_master['nickname_suggested'] != ''
        
        grouped = self.df_master .groupby(['category_precategorized','subcategory_precategorized'])
        
        summary_gruped = grouped.agg({'nickname_suggested': Series.count,
             'change_nickname': Series.sum})


        summary_gruped.reset_index(inplace = True)
        summary_gruped['Percentage of changes'] = (summary_gruped['change_nickname'] *100/ summary_gruped['nickname_suggested']).apply(floor)
        tmp  =summary_gruped['nickname_suggested']-summary_gruped['change_nickname']
        summary_gruped['Nickname accuracy percentage'] = (tmp/summary_gruped['nickname_suggested']*100).apply(floor)
        summary_gruped.columns = columns_nickname_level_2
                
        return summary_gruped[sub_cols]
        
    def summary_subcategory_lv2(self):
    
        columns_subcat_level_2 = [
            'Category',
            'Subcategory',
            'Subcategory suggested',
            'Number of changes subcategory',
            'Percentage of changes subcategory',
            'Subcategory accuracy percent'
                          ]
        grouped = self.df_master.groupby(['category_precategorized','subcategory_precategorized'])
        summary_gruped = grouped.agg({'id': Series.count,
             'change_subcategory': Series.sum})

        summary_gruped.reset_index(inplace = True)
        
        summary_gruped['Percentage of changes'] = (summary_gruped['change_subcategory'] *100/ 
          summary_gruped['id']).apply(floor)
        tmp  =summary_gruped['id']-summary_gruped['change_subcategory']
        summary_gruped['Subcategory accuracy percentage'] = (tmp/summary_gruped['id']*100).apply(floor)
        summary_gruped.columns = columns_subcat_level_2
        
        del summary_gruped['Category']
        del summary_gruped['Subcategory']

        return summary_gruped
    
    def summary_description_lv3(self):
        columns_description_level_3 = [
                            'Concept (Clean)',
                            'Classified transactions',
                            'Users (With this classified transaction)',
                            'Unclassified transactions',
                            'Frecuency (User)'
                          ]
        
        columns_red = [
                    'Concept (Clean)',
                    'Classified transactions',
                    'Users (With this classified transaction)',
                    'Frecuency (User)',
                    'Unclassified transactions',
                  ]
        
        self.df_master['classified'] = self.df_master['subcategory_precategorized'] != 'Without Classification'
        #self.df_master['unclassified'] = self.df_master['subcategory_precategorized'] == 'Without Classification'
        
        grouped = self.df_master.groupby(['clean_description','nickname_suggested','category_precategorized','subcategory_precategorized'])
        summary_gruped = grouped.agg(
            transactions=('id', Series.count),
            users=('user_id', Series.nunique),
            classified=('classified', sum))
        
        summary_gruped['frec'] = summary_gruped['transactions']/summary_gruped['users']
        summary_gruped.reset_index(inplace = True)

        del summary_gruped['nickname_suggested']
        del summary_gruped['category_precategorized']
        del summary_gruped['subcategory_precategorized']
        
        summary_gruped.columns = columns_description_level_3
        
        return summary_gruped[columns_red]

    def summary_nickname_lv3(self):
        
        cols_nickname_lv3 = ['Description clean',
                             'Nickname suggested',
                             'Number of changes nickname',
                             'Percentage of changes nickname',
                             'Nickname accuracy percent']
        
        self.df_master['classified'] = self.df_master['subcategory_precategorized'] != 'Without Classification'

        grouped = self.df_master.groupby(['clean_description','nickname_suggested','category_precategorized','subcategory_precategorized'])
        summary_gruped = grouped.agg(change_nickname=('change_nickname', Series.sum),
                                    count_=('nickname_suggested', Series.count))
        summary_gruped.reset_index(inplace = True)

        summary_gruped['Percentage of changes'] = (summary_gruped['change_nickname'] *100/ summary_gruped['count_']).apply(floor)
        tmp  =summary_gruped['count_']-summary_gruped['change_nickname']
        summary_gruped['Nickname accuracy percentage'] = (tmp/summary_gruped['count_']*100).apply(floor)
        del summary_gruped['count_']
        del summary_gruped['category_precategorized']
        del summary_gruped['subcategory_precategorized']
        
        summary_gruped.columns = cols_nickname_lv3
        
        del summary_gruped['Description clean']
        return summary_gruped
        
    def summary_subcategory_lv3(self):
        
        cols_subcat_lv3 = [
                        'Description clean',
                        'Category suggested',
                        'Subcategory suggested',
                        'Number of changes subcategory',
                        'Percentage of changes subcategory',
                        'Subcategory accuracy percent'
                        ]

        grouped = self.df_master.groupby(['clean_description','nickname_suggested','category_precategorized','subcategory_precategorized'])
        summary_gruped = grouped.agg(change_subcategory=('change_subcategory', Series.sum),
                count_=('subcategory_precategorized', Series.count))
        summary_gruped.reset_index(inplace = True)

        summary_gruped['Percentage of changes'] = (summary_gruped['change_subcategory'] *100/ 
        summary_gruped['count_']).apply(floor)
        tmp = summary_gruped['count_']-summary_gruped['change_subcategory']
        summary_gruped['Subcategory accuracy percentage'] = (tmp/summary_gruped['count_']*100)

        del summary_gruped['count_']
        del summary_gruped['nickname_suggested']
        summary_gruped.columns = cols_subcat_lv3
        del summary_gruped['Description clean']
        return summary_gruped
    
    def summary_user_lv4(self):
        cols = ['ID User',
                'Concept (Clean)',
                'Classified transactions',
                'Frecuency (User)',
                'Unclassified transactions']

        self.df_master['classified'] = self.df_master['subcategory_precategorized'] != 'Without Classification'
        self.df_master['unclassified'] = self.df_master['subcategory_precategorized'] == 'Without Classification'

        grouped = self.df_master.groupby(['user_id','clean_description','category_precategorized','subcategory_precategorized','nickname_suggested'])
        summary_gruped = grouped.agg(
                                transactions=('classified', Series.sum),
                                frecuency=('user_id', Series.nunique),
                                count_=('unclassified', Series.sum)
                                    )
        summary_gruped.reset_index(inplace = True)
        del summary_gruped['category_precategorized']
        del summary_gruped['subcategory_precategorized']
        del summary_gruped['nickname_suggested']
        summary_gruped.columns = cols

        return summary_gruped

    def summary_nickname_lv4(self):
          
        cols = ['user_id',
                'clean_description',
                'Nickname suggested',
                'Number of changes nickname',
                'Percentage of changes nickname',
                'Nickname accuracy percent'
               ]

        grouped = self.df_master.groupby(['user_id','clean_description','category_precategorized','subcategory_precategorized','nickname_suggested'])
        summary_gruped = grouped.agg(
                                change_nickname=('change_nickname', Series.sum),
                                 count_=('nickname_suggested', Series.count))

        summary_gruped.reset_index(inplace = True)

        summary_gruped['Percentage of changes'] = (summary_gruped['change_nickname'] *100/ summary_gruped['count_']).apply(floor)
        tmp  =summary_gruped['count_']-summary_gruped['change_nickname']
        
        summary_gruped['Nickname accuracy percentage'] = (tmp/summary_gruped['count_']*100).apply(floor)
        
        del summary_gruped['count_']

        
        
        del summary_gruped['category_precategorized']
        del summary_gruped['subcategory_precategorized']
        
        summary_gruped.columns = cols

        del summary_gruped['user_id']
        del summary_gruped['clean_description']
        return summary_gruped

    def summary_subcategory_lv4(self):
    
      
        cols = [
            'User ID',
            'clean_description',
            'Category suggested',
            'Subcategory suggested',
            'Number of changes subcategory',
            'Percentage of changes subcategory',
            'Subcategory accuracy percent'
            ]

        grouped = self.df_master.groupby(['user_id','clean_description','category_precategorized','subcategory_precategorized','nickname_suggested'])
        summary_gruped = grouped.agg(
                                change_nickname=('change_subcategory', Series.sum),
                                 count_=('subcategory_precategorized', Series.count))

        summary_gruped.reset_index(inplace = True)

        summary_gruped['Percentage of changes'] = (summary_gruped['change_nickname'] *100/ summary_gruped['count_']).apply(floor)
        tmp  =summary_gruped['count_']-summary_gruped['change_nickname']
        summary_gruped['Nickname accuracy percentage'] = (tmp/summary_gruped['count_']*100).apply(floor)
        del summary_gruped['count_']
        del summary_gruped['nickname_suggested']
        summary_gruped.columns = cols

        del summary_gruped['User ID']
        del summary_gruped['clean_description']

        
        return summary_gruped
    
    def summary_user_lv5(self):
        
        cols = ['ID User',
        'Nickname',
        'Transactions with this alias',
        'Percentage of transactions with this alias']
        
        grouped = self.df_master.groupby(['user_id','nickname_suggested'])
        summary_gruped = grouped.agg(
                                 count_=('subcategory_precategorized', Series.count),
                                #porcent=("id", lambda x: (max(x) - min(x)).days)    
        )

        summary_gruped.reset_index(inplace = True)
        summary_gruped['Percent of transactions with this alias'] = (summary_gruped['count_'] *100/ summary_gruped['count_'].sum()).apply(floor)
        summary_gruped.columns = cols
        
        return summary_gruped

    def summary_user_lv6(self):
        
        cols = [
            'ID User',
            'ID Transaction',
            'Full concept',
            'Clean concept',
            'Day (Of classification)',
            'Hr (Classification)'
            'Bank',
            'Nickname suggested',
            'Change',
            'User alias'
            ]

        cols = ['user_id',
                'id',
                'description',
                'clean_description',
                'bank',
                'classified_at',
                'nickname_suggested',
                'change_nickname',
                'nickname_custom'
               ]
        
        sub_cols =['user_id',
                'id',
                'description',
                'clean_description',
                'bank']
        sub_cols_ = ['nickname_suggested',
                'change_nickname',
                'nickname_custom']
        
        df_final = self.df_master[cols]


        
        cols = ['Day (Of classification)','Hr (Classification)']        

        #df_final['classified_at']= df_final['classified_at'].apply(lambda x: str(x))
        
        df_final['classified_at'] = self.df_master['updated_at']
        
        df_final['classified_at'] = df_final['classified_at'].apply(lambda x: str(x))


        df_tmp = df_final.classified_at.str.split(pat = ' ', expand = True)

        df_tmp.columns = cols
        
        del df_final['classified_at']
        
        
        
        return  pd.concat([df_final[sub_cols], df_tmp, df_final[sub_cols_]], axis=1)
    
    def generate_cubes(self):
        #-------------------------------------------------<Level One>
        df_ =  concat([self.summary_category_lv1(),
                       self.summary_nickname_lv1(),
                       self.summary_subcategory_lv1()]
                      , axis=1)
        df_['from_date'] = self.from_date
        df_['to_date'] = self.to_date
        df_.reset_index(drop=True, inplace=True)
        self.CMongo.con_db["LevelOne."+ self.from_date.strftime("%b") +".cube"].insert_many(df_.to_dict(orient='records'))
        #-------------------------------------------------<Level Two>
        df_ =  concat([self.summary_category_lv2(),
                       self.summary_nickname_lv2(),
                       self.summary_subcategory_lv2()],
                      axis = 1)
        df_['from_date'] = self.from_date
        df_['to_date'] = self.to_date
        
        df_.reset_index(drop=True, inplace=True)
        self.CMongo.con_db["LevelTwo."+ self.from_date.strftime("%b") +".cube"].insert_many(df_.to_dict(orient='records'))
        #-------------------------------------------------<Level Three>
        df_ = concat([self.summary_description_lv3(),
                      self.summary_nickname_lv3(),
                      self.summary_subcategory_lv3()],
                     axis = 1)
        df_['from_date'] = self.from_date
        df_['to_date'] = self.to_date
        df_.reset_index(drop=True, inplace=True)
        self.CMongo.con_db["LevelThree."+ self.from_date.strftime("%b") +".cube"].insert_many(df_.to_dict(orient='records'))
        #-------------------------------------------------<Level Four>
        df_ =  concat([self.summary_user_lv4(),
                       self.summary_nickname_lv4(),
                       self.summary_subcategory_lv4()],
                      axis = 1)
        df_['from_date'] = self.from_date
        df_['to_date'] = self.to_date
        df_.reset_index(drop=True, inplace=True)
        self.CMongo.con_db["LevelFour."+ self.from_date.strftime("%b") +".cube"].insert_many(df_.to_dict(orient='records'))
        #-------------------------------------------------<Level Five>
        df_ = self.summary_user_lv5()
        df_['from_date'] = self.from_date
        df_['to_date'] = self.to_date
        df_.reset_index(drop=True, inplace=True)
        self.CMongo.con_db["LevelFive."+ self.from_date.strftime("%b") +".cube"].insert_many(df_.to_dict(orient='records'))
        #-------------------------------------------------<Level Six>
        df_ = self.summary_user_lv6()
        df_['from_date'] = self.from_date
        df_['to_date'] = self.to_date
        df_.reset_index(drop=True, inplace=True)
        self.CMongo.con_db["LevelSix."+ self.from_date.strftime("%b") +".cube"].insert_many(df_.to_dict(orient='records'))

    def run(self):
        self.summary_master()
        #self.trasform_datetime()
        self.remove_conflicts()
        self.add_new_data()
        self.generate_cubes()
        self.CMongo.close_con()




from logging import getLogger
import traceback
from os import getenv
from botocore.exceptions import ClientError
import sqlalchemy as sa
from sqlalchemy.engine.url import URL
from sqlalchemy import orm as sa_orm
from math import ceil
from nanoid import generate
import sqlalchemy as sa
from sqlalchemy.engine.url import URL
from klp_commons.log.structuredMessage import StructuredMessage
from sqlalchemy import orm as sa_orm
from datetime import datetime
from pandas import read_sql_query
from sqlalchemy import text
from datetime import datetime,timedelta
from pandas import DataFrame
from pandas import concat  
#import uuid
#import random

fmt = StructuredMessage
message = 'redshiftController'


class ControllerRedShift:
    def __init__(self):

        self.error = False
        self.url_string= None
        self.metadata = None
        self.session = None
        self.engine = None
        self.region_name = None
        self.driver = 'redshift+redshift_connector'
        self.REDSHIFT_DATABASE = getenv("REDSHIFT_DATABASE")
        self.REDSHIFT_USER = getenv("REDSHIFT_USER")
        self.REDSHIFT_PASSWORD = getenv("REDSHIFT_PASSWORD")
        self.REDSHIFT_HOST = getenv("REDSHIFT_HOST")
        self.REDSHIFT_PORT = getenv("REDSHIFT_PORT")
        self.prefix = getenv("NODE_ENV") + '_'
        self.date_format= '%Y-%m-%dT%H:%M:%S'

       
        # set self.url_string
        self.get_url()

        # module_name.TopicKafka_name
        self.logger = getLogger(message)

        # Init ()
        self.logger.info("donde init method of ControllerRedShift ")

        """
        
        engine = sa.create_engine(url)
        Session = sa_orm.sessionmaker()
        Session.configure(bind=engine)
        session = Session()

        # Define Session-based Metadata
        metadata = sa.MetaData(bind=session.bind)

        # table_name
        # metadata
        # engine

        

        
        """
          
    def get_url(self):
        
        # build the sqlalchemy URL
        self.url_string = URL.create(
        drivername=self.driver,
        host=self.REDSHIFT_HOST,# Amazon Redshift host
        port=self.REDSHIFT_PORT, # Amazon Redshift port
        database=self.REDSHIFT_DATABASE, # Amazon Redshift database
        username=self.REDSHIFT_USER, # Amazon Redshift username
        password=self.REDSHIFT_PASSWORD # Amazon Redshift password
        )

    def get_conection(self) -> None:

        self.logger.info("call get_conection() method of ControllerRedShift ")

        try:
                
                self.engine = sa.create_engine(self.url_string)
                Session = sa_orm.sessionmaker()
                Session.configure(bind=self.engine)
                self.session = Session()
                # Define Session-based Metadata
                self.metadata = sa.MetaData(bind=self.session.bind)
                
        except ClientError as err:
            self.logger.error('error  get_conection RedShift')
            raise TypeError('error  get_conection RedShift')

        self.logger.info(" Done get_conection() method of RedShift")

    def close_conection(self):          
        self.session.close() 
        self.engine.dispose()
        
    def insert_dim_user(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_user (user_uid,postcode)
                        VALUES (%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['user_uid'],
                        dict_values['postcode']
                                )
            cur.execute(redshift_insert_query, record_to_insert)
            
    def insert_dim_account(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_account (account_uid,
                                                    bank_name,
                                                    account_type,
                                                    bank_type)
                        VALUES (%s,%s,%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['account_uid'],
                        dict_values['bank_name'],
                        dict_values['account_type'],
                        dict_values['bank_type']
                                )
            cur.execute(redshift_insert_query, record_to_insert)      
            
    def insert_dim_dates(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_dates (
                                                transaction_date,
                                                day_of_month,
                                                month_of_year,
                                                month_name,
                                                year,
                                                week_of_month,
                                                quarter_of_year
                                                )
                        VALUES (%s,%s,%s,%s,%s,%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['transaction_date'],
                        dict_values['day_of_month'],
                        dict_values['month_of_year'],
                        dict_values['month_name'],
                        dict_values['year'],
                        dict_values['week_of_month'],
                        dict_values['quarter_of_year']
                                )
            cur.execute(redshift_insert_query, record_to_insert)
            
    def date_decomposition(self,transaction_date):
        
        transaction_date = datetime.strptime(transaction_date, self.date_format).date()
        year = transaction_date.year
        month = transaction_date.month
        month_name = transaction_date.strftime('%b')
        day = transaction_date.day
        week_of_month = self.week_of_month(transaction_date)
        quarter_of_year = self.quarter_of_year(transaction_date)
        
        result = {
        'transaction_date':transaction_date,
        'day_of_month':day,
        'month_of_year':month,
        'month_name':month_name,
        'year':year,
        'week_of_month':week_of_month,
        'quarter_of_year':quarter_of_year
        }

        return result 
        
        # transaction_date.dt.weekday
        #transaction_date.dt.weekday_name

    def test(self):

        sql = "SELECT * FROM staging;"
        print(read_sql_query(text(sql),self.engine))
        
    def select_is_essential(self, start='2023-03-07 00:00:00', end = '2023-03-07 23:59:59'):        

        sql = f"""SELECT
                    dim_transactions.transaction_uid,
                    dim_transactions.nickname_recommended,
                    dim_transactions.is_essential_collective,
                    dim_transactions.is_essential_custom,
                    dim_transactions.is_essential_recommended
                FROM 
                    fact_transactions, dim_transactions
                WHERE 
                    fact_transactions.updated_at 
                        BETWEEN '{start}' and '{end}';"""
        return read_sql_query(text(sql),self.engine)
    
    def select_is_membership(self, start='2023-03-07 00:00:00', end = '2023-03-07 23:59:59'):        

        sql = f"""SELECT
                    dim_transactions.transaction_uid,
                    dim_transactions.nickname_recommended,
                    dim_transactions.is_membership_collective,
                    dim_transactions.is_membership_custom,
                    dim_transactions.is_membership_recommended
                FROM 
                    fact_transactions, dim_transactions
                WHERE 
                    fact_transactions.updated_at 
                        BETWEEN '{start}' and '{end}';"""
        return read_sql_query(text(sql),self.engine)
    
    
    def select_nickname(self, start='2023-03-07 00:00:00', end = '2023-03-07 23:59:59'):        

        sql = f"""SELECT
                    dim_transactions.transaction_uid,
                    dim_transactions.clean_description,
                    dim_transactions.nickname_collective,
                    dim_transactions.nickname_custom,
                    dim_transactions.nickname_recommended
                FROM 
                    fact_transactions, dim_transactions
                WHERE 
                    fact_transactions.updated_at 
                        BETWEEN '{start}' and '{end}';"""
        return read_sql_query(text(sql),self.engine)
    
    def select_subcategory(self, start='2023-03-07 00:00:00', end = '2023-03-07 23:59:59'):        

        sql = f"""SELECT
                    dim_transactions.transaction_uid,
                    dim_transactions.nickname_recommended,
                    dim_transactions.subcategory_uid_collective,
                    dim_transactions.subcategory_uid_custom,
                    dim_transactions.subcategory_uid_recommended
                FROM 
                    fact_transactions, dim_transactions
                WHERE 
                    fact_transactions.updated_at 
                        BETWEEN '{start}' and '{end}';"""
        return read_sql_query(text(sql),self.engine)
    
    
    def exists(self, table_name,column_name ,value ):

        sql = "SELECT "+ column_name +" FROM " + table_name + " WHERE " + column_name + " = '"+ value +"';"
        
        if read_sql_query(text(sql),self.engine).shape[0] != 0:
            return True
        else : return False

    def week_of_month(self,dt):
        """ Returns the week of the month for the specified date.
        """

        first_day = dt.replace(day=1)

        dom = dt.day
        adjusted_dom = dom + first_day.weekday()

        return int(ceil(adjusted_dom/7.0))

    def quarter_of_year(self,dt):

        return ceil(dt.month/3.)

    def mapping_dict_(self,dict_values):
        return {
            'approved_asset' : dict_values['approved_asset'],
            'approved_business' : dict_values['approved_business'],
            'approved_category' : dict_values['approved_category'],
            'approved_frequency' : dict_values['approved_frequency'],
            'approved_income_source' : dict_values['approved_income_source'],
            'approved_is_essential' : dict_values['approved_is_essential'],
            'approved_is_loan_received' : dict_values['approved_is_loan_received'],
            'approved_is_loan_requested' : dict_values['approved_is_loan_requested'],
            'approved_is_membership' : dict_values['approved_is_membership'],
            'approved_nickname' : dict_values['approved_nickname'],
            'approved_person' : dict_values['approved_person'],
            'approved_special_moment' : dict_values['approved_special_moment'],
            'approved_subcategory' : dict_values['approved_subcategory'],
            'asset_uid_collective' : dict_values['asset_uid_def_collect'],
            'asset_uid_custom' : dict_values['asset_uid_def_usr'],
            'asset_uid_recommended' : dict_values['asset_uid_def_algo'],
            'business_uid_collective' : dict_values['business_uid_def_collect'],
            'business_uid_custom' : dict_values['business_uid_def_usr'],
            'business_uid_recommended' : dict_values['business_uid_def_algo'],
            'category_custom_type' : dict_values['category_custom_type'],
            'category_uid_collective' : dict_values['category_uid_def_collect'],
            'category_uid_custom' : dict_values['category_uid_def_usr'],
            'category_uid_recommended' : dict_values['category_uid_def_algo'],
            'classified_date' : dict_values['classified_date'],
            'clean_description' : dict_values['clean_description'],
            'country_code' : dict_values['country_code'],
            'created_holomogation_date' : dict_values['created_holomogation_date'],
            'flow_type' : dict_values['flow_type'],
            'frequency_uid_collective' : dict_values['frequency_uid_def_collect'],
            'frequency_uid_custom' : dict_values['frequency_uid_def_usr'],
            'frequency_uid_recommended' : dict_values['frequency_uid_def_algo'],
            'income_source_uid_collective' : dict_values['income_source_uid_def_collect'],
            'income_source_uid_custom' : dict_values['income_source_uid_def_usr'],
            'income_source_uid_recommended' : dict_values['income_source_uid_def_algo'],
            'is_essential_collective' : dict_values['is_essential_def_collect'],
            'is_essential_custom' : dict_values['is_essential_def_usr'],
            'is_essential_recommended' : dict_values['is_essential_def_algo'],
            'is_loan_received_collective' : dict_values['is_loan_received_def_collect'],
            'is_loan_received_custom' : dict_values['is_loan_received_def_usr'],
            'is_loan_received_recommended' : dict_values['is_loan_received_def_algo'],
            'is_loan_requested_collective' : dict_values['is_loan_requested_def_collect'],
            'is_loan_requested_custom' : dict_values['is_loan_requested_def_usr'],
            'is_loan_requested_recommended' : dict_values['is_loan_requested_def_algo'],
            'is_membership_collective' : dict_values['is_membership_def_collect'],
            'is_membership_custom' : dict_values['is_membership_def_usr'],
            'is_membership_recommended' : dict_values['is_membership_def_algo'],
            'language_code' : dict_values['language_code'],
            'message_type' : dict_values['message_type'],
            'model_classification_version' : dict_values['model_classification_version'],
            'model_collective_version' : dict_values['model_collective_version'],
            'model_nickname_version' : dict_values['model_nickname_version'],
            'model_nlp_version' : dict_values['model_nlp_version'],
            'nickname_collective' : dict_values['nickname_collect'],
            'nickname_custom' : dict_values['nickname_usr'],
            'nickname_recommended' : dict_values['nickname_algo'],
            'original_created_date' : dict_values['original_created_date'],
            'person_uid_collective' : dict_values['person_uid_def_collect'],
            'person_uid_custom' : dict_values['person_uid_def_usr'],
            'person_uid_recommended' : dict_values['person_uid_def_algo'],
            'source_description' : dict_values['source_description'],
            'special_moment_uid_collective' : dict_values['special_moment_uid_def_collect'],
            'special_moment_uid_custom' : dict_values['special_moment_uid_def_usr'],
            'special_moment_uid_recommended' : dict_values['special_moment_uid_def_algo'],
            'subcategory_custom_type' : dict_values['subcategory_custom_type'],
            'subcategory_uid_collective' : dict_values['subcategory_uid_def_collect'],
            'subcategory_uid_custom' : dict_values['subcategory_uid_def_usr'],
            'subcategory_uid_recommended' : dict_values['subcategory_uid_def_algo'],
            'transaction_uid' : dict_values['transaction_uid'],
            'updated_date' : dict_values['updated_at'],
            'validated' : dict_values['validated'],
            'validated_date' : dict_values['creted_at']                    
        }

    def insert_dim_transaction(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO dim_transactions (
                        approved_asset,
                        approved_business,
                        approved_category,
                        approved_frequency,
                        approved_income_source,
                        approved_is_essential,
                        approved_is_loan_received,
                        approved_is_loan_requested,
                        approved_is_membership,
                        approved_nickname,
                        approved_person,
                        approved_special_moment,
                        approved_subcategory,
                        asset_uid_collective,
                        asset_uid_custom,
                        asset_uid_recommended,
                        business_uid_collective,
                        business_uid_custom,
                        business_uid_recommended,
                        category_custom_type,
                        category_uid_collective,
                        category_uid_custom,
                        category_uid_recommended,
                        classified_date,
                        clean_description,
                        country_code,
                        created_holomogation_date,
                        flow_type,
                        frequency_uid_collective,
                        frequency_uid_custom,
                        frequency_uid_recommended,
                        income_source_uid_collective,
                        income_source_uid_custom,
                        income_source_uid_recommended,
                        is_essential_collective,
                        is_essential_custom,
                        is_essential_recommended,
                        is_loan_received_collective,
                        is_loan_received_custom,
                        is_loan_received_recommended,
                        is_loan_requested_collective,
                        is_loan_requested_custom,
                        is_loan_requested_recommended,
                        is_membership_collective,
                        is_membership_custom,
                        is_membership_recommended,
                        language_code,
                        message_type,
                        model_classification_version,
                        model_collective_version,
                        model_nickname_version,
                        model_nlp_version,
                        nickname_collective,
                        nickname_custom,
                        nickname_recommended,
                        original_created_date,
                        person_uid_collective,
                        person_uid_custom,
                        person_uid_recommended,
                        source_description,
                        special_moment_uid_collective,
                        special_moment_uid_custom,
                        special_moment_uid_recommended,
                        subcategory_custom_type,
                        subcategory_uid_collective,
                        subcategory_uid_custom,
                        subcategory_uid_recommended,
                        transaction_uid,
                        updated_date,
                        validated,
                        validated_date
                                )
                        VALUES (
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s
                                )
                                    '''
            record_to_insert = (
                            dict_values['approved_asset'],
                            dict_values['approved_business'],
                            dict_values['approved_category'],
                            dict_values['approved_frequency'],
                            dict_values['approved_income_source'],
                            dict_values['approved_is_essential'],
                            dict_values['approved_is_loan_received'],
                            dict_values['approved_is_loan_requested'],
                            dict_values['approved_is_membership'],
                            dict_values['approved_nickname'],
                            dict_values['approved_person'],
                            dict_values['approved_special_moment'],
                            dict_values['approved_subcategory'],
                            dict_values['asset_uid_def_collect'],
                            dict_values['asset_uid_def_usr'],
                            dict_values['asset_uid_def_algo'],
                            dict_values['business_uid_def_collect'],
                            dict_values['business_uid_def_usr'],
                            dict_values['business_uid_def_algo'],
                            dict_values['category_custom_type'],
                            dict_values['cat_uid_def_collect'],
                            dict_values['cat_uid_def_usr'],
                            dict_values['cat_uid_def_algo'],
                            dict_values['classified_at'],
                            dict_values['clean_description'],
                            dict_values['country_code'],
                            dict_values['created_at_homologation'],
                            dict_values['flow_type'],
                            dict_values['frequency_uid_def_collect'],
                            dict_values['frequency_uid_def_usr'],
                            dict_values['frequency_uid_def_algo'],
                            dict_values['income_source_uid_def_collect'],
                            dict_values['income_source_uid_def_usr'],
                            dict_values['income_source_uid_def_algo'],
                            dict_values['is_essential_def_collect'],
                            dict_values['is_essential_def_usr'],
                            dict_values['is_essential_def_algo'],
                            dict_values['is_loan_received_def_collect'],
                            dict_values['is_loan_received_def_usr'],
                            dict_values['is_loan_received_def_algo'],
                            dict_values['is_loan_requested_def_collect'],
                            dict_values['is_loan_requested_def_usr'],
                            dict_values['is_loan_requested_def_algo'],
                            dict_values['is_membership_def_collect'],
                            dict_values['is_membership_def_usr'],
                            dict_values['is_membership_def_algo'],
                            dict_values['language_code'],
                            dict_values['message_type'],
                            dict_values['model_classification_version'],
                            dict_values['model_collective_version'],
                            dict_values['model_nickname_version'],
                            dict_values['model_nlp_version'],
                            dict_values['nickname_collect'],
                            dict_values['nickname_usr'],
                            dict_values['nickname_algo'],
                            dict_values['original_created_date'],
                            dict_values['person_uid_def_collect'],
                            dict_values['person_uid_def_usr'],
                            dict_values['person_uid_def_algo'],
                            dict_values['source_description'],
                            dict_values['special_moment_uid_def_collect'],
                            dict_values['special_moment_uid_def_usr'],
                            dict_values['special_moment_uid_def_algo'],
                            dict_values['subcategory_custom_type'],
                            dict_values['subcat_uid_def_collect'],
                            dict_values['subcat_uid_def_usr'],
                            dict_values['subcat_uid_def_algo'],
                            dict_values['transaction_uid'],
                            dict_values['updated_at'],
                            dict_values['validated'],
                            dict_values['created_at']
                            )
            cur.execute(redshift_insert_query, record_to_insert)
            
    def insert_fact_transactions(self,dict_values):
        with self.engine.connect() as cur:
            redshift_insert_query = '''
                        INSERT INTO fact_transactions (
                                                transaction_date,
                                                transaction_uid,
                                                user_uid,
                                                account_uid
                                                )
                        VALUES (%s,%s,%s,%s)
                                    '''
            record_to_insert = (
                        dict_values['transaction_date'],
                        dict_values['transaction_uid'],
                        dict_values['user_uid'],
                        dict_values['account_uid']
                                )
            cur.execute(redshift_insert_query, record_to_insert)
            
    def insert_(self):
        """
        INSERT INTO target
        SELECT * FROM temp
        WHERE temp.primary_key NOT IN (SELECT DISTINCT primary_key FROM target)
        """

    def creted_table(self,table_name = 'staging'):
        

        RedshiftDBTable = sa.Table(
                                table_name,
                                self.metadata,
                                sa.Column('session_id', sa.VARCHAR(80)),
                                sa.Column('click_region', sa.VARCHAR(100)),
                                sa.Column('product_id', sa.VARCHAR(40)),
                                redshift_diststyle='KEY',
                                redshift_distkey='session_id',
                                redshift_sortkey='product_id'
                                    )

        # Drop the table if it already exists
        if sa.inspect(self.engine).has_table(table_name):
            RedshiftDBTable.drop(bind=self.engine)

        # Create the table (execute the "CREATE TABLE" SQL statement for "")
        RedshiftDBTable.create(bind=self.engine)
        self.session.commit()

        
    def get_date_range(self,num_days = 30):
        # today 
        to_date = datetime.utcnow()
        from_date = to_date  - timedelta(days=num_days)

        return from_date, to_date


    def get_collective_is_essential(self):

        [from_date, to_date ]= self.get_date_range()

        self.get_conection()
        df_is_essential = self.select_is_essential(start = from_date, end = to_date).dropna(subset=['nickname_recommended'])

        df_is_essential = df_is_essential[df_is_essential.nickname_recommended != "No Suggestion"]
        cols = [
                'is_essential_collective',
                'is_essential_custom',
                'is_essential_recommended'
                ]

        list_cols = []

        for col in cols :
            df_tmp = df_is_essential[['transaction_uid','nickname_recommended', col]]
            df_tmp.rename(columns={col: "is_essential"},inplace=True)
            list_cols.append(df_tmp)

        result = concat(list_cols,axis=0)
        result.dropna(inplace = True)
        result.drop_duplicates(subset=['transaction_uid'], keep='last',inplace = True)
        result= DataFrame(result.groupby(['nickname_recommended','is_essential']).count()['transaction_uid'] \
                                   .groupby(['nickname_recommended','is_essential']).agg('max'))

        return result.reset_index().drop_duplicates(subset='nickname_recommended', keep='first').set_index('nickname_recommended')['is_essential'].to_dict()


    def get_collective_is_membership(self):

        [from_date, to_date ]= self.get_date_range()

        self.get_conection()
        df_is_membership = self.select_is_membership(start = from_date, end = to_date).dropna(subset=['nickname_recommended'])

        df_is_membership = df_is_membership[df_is_membership.nickname_recommended != "No Suggestion"]
        cols = [
                'is_membership_collective',
                'is_membership_custom',
                'is_membership_recommended'
                ]

        list_cols = []

        for col in cols :
            df_tmp = df_is_membership[['transaction_uid','nickname_recommended', col]]
            df_tmp.rename(columns={col: "is_membership"},inplace=True)
            list_cols.append(df_tmp)

        result = concat(list_cols,axis=0)
        result.dropna(inplace = True)
        result.drop_duplicates(subset=['transaction_uid'], keep='last',inplace = True)
        result= DataFrame(result.groupby(['nickname_recommended','is_membership']).count()['transaction_uid'] \
                                   .groupby(['nickname_recommended','is_membership']).agg('max'))

        return result.reset_index().drop_duplicates(subset='nickname_recommended', keep='first').set_index('nickname_recommended')['is_membership'].to_dict()



    def get_collective_nickname(self):

        [from_date, to_date ]= self.get_date_range()

        self.get_conection()
        df_nickname = self.select_nickname(start = from_date, end = to_date).dropna(subset=['clean_description'])

        df_nickname = df_nickname[df_nickname.nickname_recommended != "No Suggestion"]
        
        cols = [
                'nickname_collective',
                'nickname_custom',
                'nickname_recommended'
                ]

        list_cols = []

        for col in cols :
            df_tmp = df_nickname[['transaction_uid','clean_description', col]]
            df_tmp.rename(columns={col: "nickname"},inplace=True)
            list_cols.append(df_tmp)

        result = concat(list_cols,axis=0)
        result.dropna(inplace = True)
        result.drop_duplicates(subset=['transaction_uid'], keep='last',inplace = True)
        result= DataFrame(result.groupby(['clean_description','nickname']).count()['transaction_uid'] \
                                   .groupby(['clean_description','nickname']).agg('max'))

        return result.reset_index().drop_duplicates(subset='clean_description', keep='first').set_index('clean_description')['nickname'].to_dict()



    def get_collective_subcategory(self):

        [from_date, to_date ]= self.get_date_range()

        self.get_conection()
        df_subcategory = self.select_subcategory(start = from_date, end = to_date).dropna(subset=['nickname_recommended'])

        df_subcategory = df_subcategory[df_subcategory.nickname_recommended != "No Suggestion"]
        
        cols = [
            'subcategory_uid_collective',
            'subcategory_uid_custom',
            'subcategory_uid_recommended'
                ]

        list_cols = []

        for col in cols :
            df_tmp = df_subcategory[['transaction_uid','nickname_recommended', col]]
            df_tmp.rename(columns={col: "subcategory_uid"},inplace=True)
            list_cols.append(df_tmp)

        result = concat(list_cols,axis=0)
        result.dropna(inplace = True)
        result.drop_duplicates(subset=['transaction_uid'], keep='last',inplace = True)
        result= DataFrame(result.groupby(['nickname_recommended','subcategory_uid']).count()['transaction_uid'] \
                                   .groupby(['nickname_recommended','subcategory_uid']).agg('max'))

        return result.reset_index().drop_duplicates(subset='nickname_recommended', keep='first').set_index('nickname_recommended')['subcategory_uid'].to_dict()


from json import dumps


class StructuredMessage:
    '''
    Clase que define una strtura JSON para la información
    que se envía a los logs

    '''

    def __init__(self, message, /, **kwargs):
        self.message = message
        self.kwargs = kwargs

    def __str__(self):
        return '%s >>> %s' % (self.message, dumps(self.kwargs))



# ------------------------------------<Input>

null = None
false = False

# keys :
# fix 
# hardcode
# bug -> atender de inmediato
# refactorin
from distutils.log import ERROR

class CodeRalation:
    def __init__(self) -> None:
        pass

class TypeTransEnrichment:
    def __init__(self):
        self.NAME_SPACE_TRANS = {
            # delegar a MS-Report
            'user_uid': None,
            # delegar a MS-Report
            'transaction_uid': None,
            'message-type' : None,
            'offset': None,
            'data': None,
            'dynamo': None
            }

    def get(self):
        return self.NAME_SPACE_TRANS

class TypeTransHomologated:
    def __init__(self):

        self.USER = dict()
        self.DYNAMO = dict()
        self.POSTGRES = dict()
        self.CASE = dict()
        self.NLP = dict()
        self.MODEL = dict()

        self.USER['transaction_uid'] = None
        self.USER['user_uid'] = None
        self.USER['account_uid'] = None
        self.USER['description'] = None
        self.USER['amount'] = None
        self.USER['raw'] = None
        # -------------------------------------<DynamoDB>
        self.DYNAMO['keys_user'] = None
        self.DYNAMO['keys_multicategory'] = None
        self.DYNAMO['details'] = None

        # --------------------------------------<Posgrest>
        self.POSTGRES['r_query_pgdb'] = None
        self.POSTGRES['tup_r_cat'] = None
        self.POSTGRES['tup_r_subcat'] = None

        # -------------------------------------<Hashing aka description embedding>
        # Model context
        # aka hashig embedding
        # ndarray to str for hashig
        self.MODEL['features_string'] = None
        self.MODEL['threshold'] = False
        self.MODEL['predict_value'] = None
        self.MODEL['info_subcategory'] = None
        self.MODEL['hash'] = None
        self.MODEL['collective'] = None
        self.MODEL['model_uid'] = None
        self.MODEL['version_model'] = None
        # ------------------------------------<Variables de NLP>
        # reset in reset_values_class


        # ------------------------------------<Case>
        self.CASE['options_merchant_names'] = None # lista de posibles merchant-namesl7f
        self.CASE['case'] = None # standar|custom|new|multicategory
        self.CASE['single_subcat'] = None #single or multicategory
        self.CASE['merchant_uid'] = None
        self.CASE['categorization'] = None # response de MSCategorization
        self.CASE['categorized'] = None #for re-categorized :false|flase

        
    # ------------------------------------<dict all context vars>

        self.NAME_SPACE_TRANS = {
                            # delegar a MS-Report
                            'user_uid':self.USER['user_uid'],
                            # delegar a MS-Report
                            'transaction_uid':self.USER['transaction_uid'],
                            'USER': self.USER,
                            'DYNAMO': self.DYNAMO,
                            'CASE':self.CASE,
                            'NLP':self.NLP,
                            'MODEL':self.MODEL}

    def get(self):
        return self.NAME_SPACE_TRANS

class ErrorProcess:
    """
    Mensaje de error enviado al type : 'transaction-process-error' 
    lo utiliza Backend para reenviar mensajes qe tuvieron error. 
    """
    def __init__(self):
        self.ERROR_ = dict()
        self.ERROR_['type'] = 'transaction-process-error' 
        self.ERROR_['data'] = {}
        self.ERROR_['data']['institution'] = None
        self.ERROR_['data']['user_uid'] = None
    def get(self):
        return self.ERROR_

class Error:
    """
    Mensaje de error enviado al type :'report-error-ms-categorization'
    lo utiliza reporting (MS monitoring) para almacenar en Mongo los 
    errores y sus detalles
    """
    def __init__(self):
        self.item = dict()
        self.item['type'] = 'report-error-ms-categorization'
        self.item['data'] = None
        self.item['offset'] = None
        self.item['dt'] = None
        self.item['info_mixin'] = None
        self.item['error_type'] = None
        self.item['error_value'] = None
        self.item['traceback_info'] = None
        self.item['trans_context'] = None
    def get(self):
        return self.item

class JSONErrorTraceback:
    """
    JSON que almacena detalles de errores y los envía a los Logs y a Sentry
    """
    def __init__(self)->None:
        self.Item =  {
                'type': None,
                'error_type': None,
                'error_value': None,
                'traceback_info': None,
                'trans_context': None,
                'details' : JSONErrorDetails().get()
                }

    def get(self):
        return self.Item

class JSONErrorDetails:
    """
    JSON que almacena detalles de errores y los envía a los Logs y a Sentry
    """
    def __init__(self)->None:
        self.Item =  {
            'uid': None,
            'user_uid': None,
            'transaction_uid': None,
            'level_processing': None,
            'description_embedding': None,
            'DYNAMO':None
				}

    def get(self):
        return self.Item

class MsgCatCompleted :
    def __init__(self):
        self.COMPLETED = dict()
        self.COMPLETED['type'] = 'transaction-categorized-completed'
        self.COMPLETED['data']= {}
        self.COMPLETED['data']['institution'] = None
        self.COMPLETED['data']['user_uid'] = None
    def get(self):
        return self.COMPLETED

class TypeDB:
    def __init__(self):

        self.DYNAMO = dict()
        self.POSTGRES = dict()
        self.LANG = dict()

        # ------------------------------<languaje>
        self.LANG['lang_code_trans'] = None
        self.LANG['lang_code_app'] = None
        self.LANG['country_code'] = None
        # -------------------------------------<DynamoDB>
        self.DYNAMO['keys_user'] = None
        self.DYNAMO['keys_multicategory'] = None
        self.DYNAMO['details'] = None

        # --------------------------------------<Posgrest>
        self.POSTGRES['r_query_pgdb'] = None
        self.POSTGRES['tup_r_cat'] = None
        self.POSTGRES['tup_r_subcat'] = None

        self.NAME_SPACE_DB = {'DYNAMO': self.DYNAMO,
                              'POSTGRES': self.POSTGRES,
                              'LANG': self.LANG}

    def get(self):
        return self.NAME_SPACE_DB

class CategorizationObject:
    """
    Estructura del objeto categorization 
    dentro de los mensajes de Kafka 
    topic : "transaction"
    type : "transaction-categorized"

    """
    def __init__(self)->None:
        self.Item =  {
            "subcategory_uid": None, #ID subcategoría
            "category_uid": None, # ID categoría
            "merchant_name": None, # nombre del comerción
            "description": None, # Nickname del comerción.
            "clean_description": None, # Descripción limpia
            "kw_group": None, # Keyword asociado al grupo(segmento) al que pertence la transacción
            "asset_uid": None,  # 
            "business_uid": None, #
            "person_uid": None, # ID de la persona asociada a la transacción
            "special_moment_uid": None, # ID de el momento especial asociado a la transacción
            "frequency_uid": None, # ID de la frecuencia asociada a la transacción
            "is_essential": None, # True| False  si el gasto es escencial
            "is_membership": None, # True| False si el gasto pertence a un menbresía
            "income_source_uid" : None,
            "is_loan_received" : None,
            "is_loan_requested": None,
            "collective_subcategory_uid" : None,
            "collective_description" : None,
            "collective_is_essential" : None,
            "collective_is_membership" : None
			}

    def get(self):
        return self.Item

class Message:
    def __init__(self) -> None:

        self.msj = {
            'data': [],
            'type':None
                    }
    def get(self):
        return self.msj

class MsgReportStatus :
    def __init__(self):
        self.item = dict()
        self.item['type'] = 'report-message-ok'
        self.item['data']= {}
        self.item['data']['offset'] = None
        self.item['data']['type'] = None
        self.item['data']['status'] = 'ok'

    def get(self):
        return self.item

# RedShift DataWare House
class DimTransactions:
    """
    For RedShift Dimension Transactions
    """
    def __init__(self) -> None:
        self.Item = {
        "transaction_uid": None,
        "language_code": None,
        "country_code": None,
        "flow_type": None,
        "message_type": None,
        "validated": None,
        "source_description": None,
        "clean_description": None,
        "model_nlp_version": None,
        "model_classification_version": None,
        "model_nickname_version": None,
        "model_collective_version": None,
        "approved_asset": None,
        "approved_business": None,
        "approved_category": None,
        "approved_frequency": None,
        "approved_is_essential": None,
        "approved_is_membership": None,
        "approved_nickname": None,
        "approved_person": None,
        "approved_special_moment": None,
        "approved_subcategory": None,
        "approved_is_loan_received": None,
        "approved_is_loan_requested": None,
        "approved_income_source": None,
        "is_loan_requested_recommended": None,
        "is_membership_recommended": None,
        "subcategory_uid_recommended": None,
        "category_uid_recommended": None,
        "special_moment_uid_recommended": None,
        "is_essential_recommended": None,
        "frequency_uid_recommended": None,
        "asset_uid_recommended": None,
        "person_uid_recommended": None,
        "business_uid_recommended": None,
        "income_source_uid_recommended": None,
        "is_loan_received_recommended": None,
        "nickname_recommended": None,
        "subcategory_custom_type": None,
        "category_custom_type": None,
        "is_essential_custom": None,
        "is_membership_custom": None,
        "person_uid_custom": None,
        "is_loan_requested_custom": None,
        "special_moment_uid_custom": None,
        "business_uid_custom": None,
        "asset_uid_custom": None,
        "subcategory_uid_custom": None,
        "frequency_uid_custom": None,
        "category_uid_custom": None,
        "income_source_uid_custom": None,
        "is_loan_received_custom": None,
        "nickname_custom": None,
        "category_uid_collective": None,
        "subcategory_uid_collective": None,
        "nickname_collective": None,
        "income_source_uid_collective": None,
        "is_loan_requested_collective": None,
        "is_essential_collective": None,
        "is_membership_collective": None,
        "frequency_uid_collective": None,
        "is_loan_received_collective": None,
        "person_uid_collective": None,
        "asset_uid_collective": None,
        "business_uid_collective": None,
        "special_moment_uid_collective": None,
        "original_created_date": None,
        "created_holomogation_date": None,
        "classified_date": None,
        "validated_date": None,
        "updated_date": None,
        "created_at": None,
        "updated_at": None,
        "deleted_at": None
            }
    def get(self):
        return self.Item

class DimAccount:
    """
    """
    def __init__(self) -> None:
        self.Item = {
    "account_uid" : None,
    "bank_name" : None,
    "account_type" : None,
    "bank_type" : None,
    "created_at" : None,
    "updated_at" : None,
    "deleted_at" : None
    }
    def get(self):
        return self.Item

class DimUser:
    """
    """
    def __init__(self) -> None:
        self.Item = {
    "user_uid" : None,
    # "postcode" : None,
    # "created_at" : None,
    # "updated_at" : None,
    # "deleted_at" : None
    }
    def get(self):
        return self.Item
    
class DimDates:
    """
    """
    def __init__(self) -> None:
        self.Item = {
    "transaction_date" : None,# validated or updated
    "day_of_month" : None,
    "month_of_year" : None,
    "month_name"  : None,
    "year" : None,
    "week_of_month" : None,
    "quarter_of_year" : None,
    # "created_at" : None,
    # "updated_at" : None,
    # "deleted_at" : None
    }
    def get(self):
        return self.Item

class FactTransactions:

    def __init__(self) -> None:
        
        self.Item = {
            "transaction_date": None,
            "transaction_uid": None,
            "user_uid": None,
            "account_uid": None
        }

    def get(self):
        return self.Item

# MongoDB History Facts
class TransactionFact:
    """
    For mongoDB 
    """
    def __init__(self) -> None:
        self.Item = {
            "_id": None,
            "user_uid": None,
            "idx": None,
            "offset": None,
            "hash": None,
            "amount": None,
            "currency": None,
            "account_institution_name": None,
            "account_institution_type": None,
            "source_category": null,
            "account_type": None,
            "account_name": None,
            "account_number": None,
            "account_bank_product_id": None,
            "account_internal_identification": None,
            "account_public_identification_value": None,
            "account_public_identification_name": None,
            "account_category": None,
            "source_merchant": None,
            "source_description": None,
            "value_date": None,
            "collected_at": None,
            "message_type": None,
            "original_created_at": None,
            "lang_code": None,
            "country_code": None,
            "created_at_homologation": None,
            "type_flow": None,
            "history-track":[]
            }
    def get(self):
        return self.Item

class PreCategorized:
    """
    """
    def __init__(self) -> None:
        self.Item = {
            "subcategory_uid": None, #ID subcategoría
            "category_uid": None, # ID categoría
            "merchant_name": None, # nombre del comerción
            "nickname": None, # Nickname del comerción.
            "clean_description": None, # Descripción limpia
            "kw_group": None, # Keyword asociado al grupo(segmento) al que pertence la transacción
            "asset_uid": None,  # 
            "business_uid": None, #
            "person_uid": None, # ID de la persona asociada a la transacción
            "special_moment_uid": None, # ID de el momento especial asociado a la transacción
            "frequency_uid": None, # ID de la frecuencia asociada a la transacción
            "is_essential": None, # True| False  si el gasto es escencial
            "is_membership": None, # True| False si el gasto pertence a un menbresía
            "income_source_uid" : None,
            "is_loan_received" : None,
            "is_loan_requested": None,
            "status": None,
            "categorized": None,
            "replica_type": None,
            "message_type": None,
            "classified_at": None,
            "model_nlp_version": None,
            "model_classification_version": None,
            "model_nickname_version": None,
            "model_collective_version": None,
        }
    def get(self):
        return self.Item
    
class ValidatedInfo:
    """
    """
    def __init__(self) -> None:
        self.Item = {
        {
     
        }

        }
    def get(self):
        return self.Item
       
class UpdatedInfo:
    """
    """
    def __init__(self) -> None:
        self.Item = {
        {
      
        }

        }
    def get(self):
        return self.Item

# DynamoDB
class DetailsUser:
    """
    Objeto  DetailsUser, el cual es un objeto anidado del super Objeto ItemUser
    """

    def __init__(self) -> None:
        self.Item = {

        "cat_uid_def_algo": None,
        "cat_uid_def_usr": None,
        "cat_uid_def_collect": None,

        "nickname_algo": None,
        "nickname_usr": None,
        "nickname_collect": None,

        "subcat_uid_def_algo": None,
        "subcat_uid_def_usr": None,
        "subcat_uid_def_collect": None,

        "frequency_uid_def_algo": None,
        "frequency_uid_def_usr": None,
        "frequency_uid_def_collect": None,

        "is_essential_def_algo": None,
        "is_essential_def_usr": None,
        "is_essential_def_collect": None,

        "is_membership_def_algo": None,
        "is_membership_def_usr": None,
        "is_membership_def_collect": None,

        "person_uid_def_usr": None,
        "person_uid_def_algo": None,
        "person_uid_def_collect": None,

        "special_moment_uid_def_algo": None,
        "special_moment_uid_def_usr": None,
        "special_moment_uid_def_collect": None,

        "business_uid_def_algo": None,
        "business_uid_def_usr": None,
        "business_uid_def_collect": None,

        "asset_uid_def_algo": None,
        "asset_uid_def_usr": None,
        "asset_uid_def_collect": None,

        "income_source_uid_def_algo": None,
        "income_source_uid_def_usr": None,
        "income_source_uid_def_collect": None,

        "is_loan_received_def_algo": None,
        "is_loan_received_def_usr": None,
        "is_loan_received_def_collect": None,

        "is_loan_requested_def_algo": None,
        "is_loan_requested_def_usr": None,
        "is_loan_requested_def_collect": None,

        "created_at": None,# type:expense-created
        "updated_at": None,# type:expense-updated
        
        "message_type": None,
        "clean_description": None,
        "merchant_name": None,
        "classified_at": None, # type:transaction-categorized
        "source_embedding": None,
        "kw_group":None

        }

    def get(self):
        return self.Item

class ItemUser:
    """
     Obtejo ItemUser, el cual representa un documento en la colección 
     User de DynamoDB en AWS.+


    """
    def __init__(self) -> None:
        self.Item = {
            "user_uid": None, #Clave de partición
            "description_embedding": None, # Clave de búsqueda

            "approved_cat": None,
            "approved_nickname": None,
            "approved_subcat": None,
            "approved_is_essential": None,
            "approved_is_membership": None,
            "approved_person": None,
            "approved_special_moment": None,
            "approved_frequency": None,
            "approved_business": None,
            "approved_asset": None,
            "approved_is_loan_received": None,
            "approved_is_loan_requested": None,
            "approved_income_source_uid": None,

            "change_asset": None,
            "change_business": None,
            "change_cat": None,
            "change_frequency": None,
            "change_income_source_uid": None,
            "change_is_essential": None,
            "change_is_loan_received": None,
            "change_is_loan_requested": None,
            "change_is_membership": None,
            "change_nickname": None,
            "change_person": None,
            "change_special_moment": None,
            "change_subcat": None,

            "validated": None,
            "language_code": None,
            "categorized": None,
            "replica_type": None,# agregar a variables *
            "message_type": None, #typo de mensaje kafka
            "flow_type": None,# inflow|outflow
            "created_at_homologation": None, # type:transacton-homologated
            
            "case_type": None,
            
            "model_nlp_version": None,
            "model_classification_version": None,
            "model_nickname_version": None,
            "model_collective_version": None,
            "details": None # NestedDict type DetailsUser()
            }
    def get(self):
        return self.Item

class ItemMulty:
    def __init__(self) -> None:

        self.Item = {
            'description_embedding': None,
            'popular_subcategories': 
                        {   
                        1: None,
                        2: None,
                        3:None
                         },
            'categories': 
                        {
                        1: None,
                        2: None,
                        3:None
                        },
            'collective_nickname': None,
            'nickname': None,
            'merch_uid': None,
            'subcat_uid': None
            }
    def get(self):
        return self.Item

class ItemTrans:
    # Tabla Transaction en DynamoDB
    def __init__(self) -> None:

        self.Item = {
            'user_uid': None,
            'transaction_uid': None,
            'hash': None
                    }
    def get(self):
        return self.Item

# Diccionarios
class Transaction_:

    def __init__(self) -> None:

        self.Item = {
            'approved_asset': None,
            'approved_business': None,
            'approved_category': None,
            'approved_frequency': None,
            'approved_income_source': None,
            'approved_is_essential': None,
            'approved_is_loan_received': None,
            'approved_is_loan_requested': None,
            'approved_is_membership': None,
            'approved_nickname': None,
            'approved_person': None,
            'approved_special_moment': None,
            'approved_subcategory': None,
            'asset_uid_def_collect': None,
            'asset_uid_def_algo': None,
            'asset_uid_def_usr': None,
            'business_uid_def_collect': None,
            'business_uid_def_algo': None,
            'business_uid_def_usr': None,
            'category_custom_type': None,
            'cat_uid_def_collect': None,
            'cat_uid_def_algo': None,
            'cat_uid_def_usr': None,
            'classified_at': None,
            'clean_description': None,
            'country_code': None,
            'created_at': None,
            'created_at_homologation': None,
            'flow_type': None,
            'frequency_uid_def_collect': None,
            'frequency_uid_def_algo': None,
            'frequency_uid_def_usr': None,
            'income_source_uid_def_collect': None,
            'income_source_uid_def_algo': None,
            'income_source_uid_def_usr': None,
            'is_essential_def_collect': None,
            'is_essential_def_algo': None,
            'is_essential_def_usr': None,
            'is_loan_received_def_collect': None,
            'is_loan_received_def_algo': None,
            'is_loan_received_def_usr': None,
            'is_loan_requested_def_collect': None,
            'is_loan_requested_def_algo': None,
            'is_loan_requested_def_usr': None,
            'is_membership_def_collect': None,
            'is_membership_def_algo': None,
            'is_membership_def_usr': None,
            'language_code': None,
            'message_type': None,
            'model_classification_version': None,
            'model_collective_version': None,
            'model_nickname_version': None,
            'model_nlp_version': None,
            'nickname_algo': None,
            'nickname_collect': None,
            'nickname_usr': None,
            'original_created_date': None,
            'person_uid_def_collect': None,
            'person_uid_def_algo': None,
            'person_uid_def_usr': None,
            'source_description': None,
            'special_moment_uid_def_collect': None,
            'special_moment_uid_def_algo': None,
            'special_moment_uid_def_usr': None,
            'subcategory_custom_type': None,
            'subcat_uid_def_usr': None,
            'subcat_uid_def_collect': None,
            'subcat_uid_def_algo': None,
            'updated_at': None,
            'validated': None,
            'transaction_uid': None,
            }
    
    def get(self):
        return self.Item


from json import loads as loads_
from confluent_kafka import Consumer as Consumer_
from kafka import KafkaConsumer as KafkaConsumer_
from kafka import TopicPartition

class Consumer:
    '''
    Schema (or event type):transaction-homologated
    topic (string): 'transaction'
    key (string): undefined
    value (string):



    '''
     
    def __init__(self, topics_list = None,
                         kafka_brokers = None,
                          group_id = None, 
                          api_version = None,
                          client_id = None,
                          heartbeat_interval_ms = None,
                          session_timeout_ms = None,
                          auto_offset_reset = None
                          ):


        self.auto_offset_reset = auto_offset_reset
        self.session_timeout_ms = session_timeout_ms
        self.heartbeat_interval_ms = heartbeat_interval_ms
        self.topics_list = topics_list
        self.group_id = group_id
        self.kafka_brokers = kafka_brokers
        self.kafka_api_version = api_version
        self._consumer = None
        self.client_id = client_id
        self.data = []
        

    def init_consumer(self):

        self._consumer = KafkaConsumer_(
            api_version= self.kafka_api_version,
            auto_offset_reset=self.auto_offset_reset,
            session_timeout_ms = self.session_timeout_ms,
            heartbeat_interval_ms = self.heartbeat_interval_ms,
            enable_auto_commit=True,
            bootstrap_servers = self.kafka_brokers,
            group_id = self.group_id,
            value_deserializer = lambda x: loads_(x.decode('utf-8')),
        )
        self._consumer.subscribe(self.topics_list)

    @property
    def consumer(self):
        return self._consumer

    @consumer.setter
    def consumer(self, value):
        if isinstance(value, KafkaConsumer_):
            self._consumer = value

    def star_read(self):
        self.receive_message()

    

    def commit_check(self,partition,KAFKA_TOPIC_NAME):
        # message.partition
        # Optionally, To check if everything went good
        return self._consumer.committed(TopicPartition(KAFKA_TOPIC_NAME,partition ))


class KafkaConsumer:
    '''
    Schema (or event type):transaction-homologated
    topic (string): 'transaction'
    key (string): undefined
    value (string):

    '''

    def __init__(self, topics_list = None,
                         kafka_brokers = None,
                          group_id = None, 
                          api_version = None,
                          heartbeat_interval_ms = None,
                          session_timeout_ms = None,
                          auto_offset_reset = None
                          ):
        
        self.heartbeat_interval_ms = heartbeat_interval_ms
        self.session_timeout_ms = session_timeout_ms
        self.auto_offset_reset = auto_offset_reset
        self.topics_list = topics_list
        self.group_id = group_id
        self.kafka_brokers = kafka_brokers
        self.kafka_api_version = api_version
        self._consumer = None
        self.client_id = None
        

    def init_consumer(self):
        
        conf = {'bootstrap.servers': self.kafka_brokers,
                'group.id': self.group_id,
                'enable.auto.commit': True,
                'auto.offset.reset': 'smallest'}
                    
        self._consumer = Consumer_(conf)

        self._consumer.subscribe(self.topics_list)

    @property
    def consumer(self):
        return self._consumer

    @consumer.setter
    def consumer(self, value):
        if isinstance(value, Consumer_):
            self._consumer = value


if __name__ == '__main__':
    pass



import base64
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

from os import getenv


class Cryptology:
    '''
    Encryotar y desemcriptar mensajes de Kafka
    '''

    # Variables de clase

    def __init__(self, pad: int = None):
        # Padding
        self.PAD = pad
        self.key = getenv('JWT_SECRET')  # Must Be 16 char for AES128

    def encrypt(self, raw):
        raw = pad(raw.encode(), self.PAD)
        cipher = AES.new(self.key.encode('utf-8'), AES.MODE_ECB)
        return base64.b64encode(cipher.encrypt(raw)).decode('utf-8', 'ignore')

    def decrypt(self, enc):
        enc = base64.b64decode(enc)
        cipher = AES.new(self.key.encode('utf-8'), AES.MODE_ECB)
        return unpad(cipher.decrypt(enc), self.PAD).decode('utf-8', 'ignore')



import copy
import logging
import regex as re
from dateutil import tz, parser
from klp_commons.datefinder.date_fragment import DateFragment
from .constants import (
    REPLACEMENTS,
    DELIMITERS_PATTERN,
    TIMEZONE_REPLACEMENTS,
    STRIP_CHARS,
    DATE_REGEX,
    ALL_GROUPS,
    RANGE_SPLIT_REGEX,
)

logger = logging.getLogger("datefinder")


class DateFinder(object):
    """
    Locates dates in a text
    """

    def __init__(self, base_date=None, first="month"):
        self.base_date = base_date
        self.dayfirst = False
        self.yearfirst = False
        if first == "day":
            self.dayfirst = True
        if first == "year":
            self.yearfirst = True

    def find_dates(self, text, source=True, index=True, strict=False):

        for date_string, indices, captures in self.extract_date_strings(text, strict=strict):

            #  [('Jan', 'January'), ('Feb', 'February'), ('Mar', 'March'), ('Apr', 'April'), 
            # ('May', 'May'), ('Jun', 'June'), ('Jul', 'July'), ('Aug', 'August'), 
            # ('Sep', 'Sept', 'September'), ('Oct', 'October'), ('Nov', 'November'), ('Dec', 'December')]
            date_string = date_string.replace('ago', 'aug').replace('julio', 'jul').replace('junio', 'jun').replace('enero', 'jan').replace('febrero', 'frb').replace('marzo', 'mar').replace('abril', 'apr').replace('mayo', 'may').replace('agosto', 'aug').replace('septiembre', 'sep').replace('octubre', 'oct').replace('noviembre', 'nov').replace('diciembre', 'dec').replace('enero', 'jan').replace('marz', 'mar')
            
            
            
            as_dt = self.parse_date_string(date_string, captures)
            
            if as_dt is None:
                ## Dateutil couldn't make heads or tails of it
                ## move on to next
                continue

            returnables = (as_dt,)
            if source:
                returnables = returnables + (date_string,)
            if index:
                returnables = returnables + (indices,)

            if len(returnables) == 1:
                returnables = returnables[0]
            yield returnables

    def _find_and_replace(self, date_string, captures):
        """
        :warning: when multiple tz matches exist the last sorted capture will trump
        :param date_string:
        :return: date_string, tz_string
        """
        # add timezones to replace
        cloned_replacements = copy.copy(REPLACEMENTS)  # don't mutate
        for tz_string in captures.get("timezones", []):
            cloned_replacements.update({tz_string: " "})

        date_string = date_string.lower()
        for key, replacement in cloned_replacements.items():
            # we really want to match all permutations of the key surrounded by whitespace chars except one
            # for example: consider the key = 'to'
            # 1. match 'to '
            # 2. match ' to'
            # 3. match ' to '
            # but never match r'(\s|)to(\s|)' which would make 'october' > 'ocber'
            # but also include delimiters, like this 'date: '
            full_match_pattern = (
                r"(^|{delimiters_pattern}){key}($|{delimiters_pattern})".format(
                    delimiters_pattern=DELIMITERS_PATTERN, key=key
                )
            )
            date_string = re.sub(
                full_match_pattern,
                replacement,
                date_string,
                flags=re.IGNORECASE,
            )

        return date_string, self._pop_tz_string(sorted(captures.get("timezones", [])))

    def _pop_tz_string(self, list_of_timezones):
        try:
            tz_string = list_of_timezones.pop()
            # make sure it's not a timezone we
            # want replaced with better abbreviation
            return TIMEZONE_REPLACEMENTS.get(tz_string, tz_string)
        except IndexError:
            return ""

    def _add_tzinfo(self, datetime_obj, tz_string):
        """
        take a naive datetime and add dateutil.tz.tzinfo object

        :param datetime_obj: naive datetime object
        :return: datetime object with tzinfo
        """
        if datetime_obj is None:
            return None

        tzinfo_match = tz.gettz(tz_string)
        return datetime_obj.replace(tzinfo=tzinfo_match)

    def parse_date_string(self, date_string, captures):
        # For well formatted string, we can already let dateutils parse them
        # otherwise self._find_and_replace method might corrupt them
        # print("date_string : ",date_string)
        try:
            as_dt = parser.parse(
                date_string,
                default=self.base_date,
                dayfirst=self.dayfirst,
                yearfirst=self.yearfirst,
            )
        except (ValueError, OverflowError):
            # replace tokens that are problematic for dateutil
            date_string, tz_string = self._find_and_replace(date_string, captures)

            ## One last sweep after removing
            date_string = date_string.strip(STRIP_CHARS)
            ## Match strings must be at least 3 characters long
            ## < 3 tends to be garbage
            if len(date_string) < 3:
                return None

            try:
                logger.debug("Parsing {0} with dateutil".format(date_string))
                as_dt = parser.parse(
                    date_string,
                    default=self.base_date,
                    dayfirst=self.dayfirst,
                    yearfirst=self.yearfirst,
                )
            except Exception as e:
                logger.debug(e)
                as_dt = None
            if tz_string:
                as_dt = self._add_tzinfo(as_dt, tz_string)
        return as_dt

    def extract_date_strings(self, text, strict=False):
        """
        Scans text for possible datetime strings and extracts them
        :param strict: Strict mode will only return dates sourced with day, month, and year
        """
        return self.extract_date_strings_inner(text, text_start=0, strict=strict)

    
    
    
    def extract_date_strings_inner(self, text, text_start=0, strict=False):
        """
        Extends extract_date_strings by text_start parameter: used in recursive calls to
        store true text coordinates in output
        """

        # Try to find ranges first
        rng = self.split_date_range(text)
        #print("ranges : ", rng)
        if rng and len(rng) > 1:
            range_strings = []
            for range_str in rng:
                range_strings.extend(
                    self.extract_date_strings_inner(
                        range_str[0], text_start=range_str[1][0], strict=strict
                    )
                )
            #print("range_strings: ",range_strings)
            for range_string in range_strings:
                yield range_string
            return
        
        tokens = self.tokenize_string(text)

        # print("tokens: " , tokens)
        items = self.merge_tokens(tokens)

        #print("items: ", items)
        for match in items:
            #print("match: ", match)

            match_str = match.match_str

            indices = (match.indices[0] + text_start, match.indices[1] + text_start)

            ## Get individual group matches
            captures = match.captures
            # time = captures.get('time')
            digits = captures.get("digits")
            # digits_modifiers = captures.get('digits_modifiers')
            # days = captures.get('days')
            months = captures.get("months")
            years = captures.get("years")
            # timezones = captures.get('timezones')
            # delimiters = captures.get('delimiters')
            # time_periods = captures.get('time_periods')
            # extra_tokens = captures.get('extra_tokens')

            if strict:
                complete = False
                if len(digits) == 3:  # 12-05-2015
                    complete = True
                elif (len(months) == 1) and (len(digits) == 2):  # 19 February 2013 year 09:10
                    complete = True
                elif (len(months) == 1) and (len(digits) == 1):  # 19 February 
                    complete = True
                elif (len(months) == 1) and (len(years) == 1):  # 2022 February 
                    complete = True
                elif (len(years) == 1) and (len(digits) == 2):  # 09/06/2018
                    complete = True

                elif ((len(years) == 1) and (len(months) == 1) and (len(digits) == 1)):  # '19th day of May, 2015'
                    complete = True

                if not complete:
                    #print("No complete format")
                    continue

            ## sanitize date string
            ## replace unhelpful whitespace characters with single whitespace
            match_str = re.sub(r"[\n\t\s\xa0]+", " ", match_str)
            # print("match_str: ", list(match_str))
            
            match_str = match_str.strip(STRIP_CHARS)
            
            #print("match_str_strip: ", list(match_str))
            # Save sanitized source string
            yield match_str, indices, captures

    def tokenize_string(self, text):
        """
        Get matches from source text. Method merge_tokens will later compose
        potential date strings out of these matches.
        :param text: source text like 'the big fight at 2p.m. mountain standard time on ufc.com'
        :return: [(match_text, match_group, {match.capturesdict()}), ...]
        """
        items = []
        
        split_text = text.split(' ')
        
        last_index = 0
        
        #print(text)
        for match in DATE_REGEX.finditer(text):
            flag_valid = True
            match_str = match.group(0)
            # print("match_str: ", match_str)

            indices = match.span(0)
            #print("indices: ", indices)

            captures = match.capturesdict()
            
            group = self.get_token_group(captures)
            
            for item in split_text:
                # print( match_str," -> ",item)
                if match_str in item  and match_str != item and  (match_str.isdigit() is False  and item.isdigit() is False):
                    # print("set '' {} by substring")
                    group = ''
                    captures = {}
            
            if match_str.isdigit():
                match_int = int(match_str)

                if len(match_str) == 2:
                    if match_int > 31:
                        # print("set '' {} by len(2)  match_int > 31")
                        group = ''
                        captures = {}
                        
                if len(match_str) == 3:
                    if match_int != 202:
                        # print("set '' {} by len(3)  !=  202")
                        group = ''
                        captures = {}
            
            
            if indices[0] > last_index :
                # print("set '' {} indices[0] > last_index")
                items.append((text[last_index : indices[0]], "", {}))
                
            items.append((match_str, group, captures))
            last_index = indices[1]
            
            
        if last_index < len(text):
            # print("set '' {} last_index < len(text)")
            items.append((text[last_index : len(text)], "", {}))
            
        return items
    
    

    def stack_delimiters(self,tokens):
        delimiters_list =[]
        
        for idx,token in enumerate(tokens):
            if 'delimiters' in token[2]:
                if len(token[2]['delimiters']) != 0:
                    delimiters_list.append((token[2]['delimiters'][0],idx))
                else : 
                    continue 
                    
            else : delimiters_list.append(('no_group',idx))
        
        return delimiters_list
                                           
    def find_delimiter(self, stack_delimit,start_index, current_index):
        
        return [tupla[0] for tupla in stack_delimit if tupla[1] > start_index and tupla[1] <= current_index ]
                                           
                                           
    def find_next_delimiters(self,tokens):
        
        for token in tokens:
            if 'delimiters' in token[2]:
                if len(token[2]['delimiters']) != 0:
                    return token[2]['delimiters'][0]
                else : continue 
            else : return 'no_group'
        
        
        
                                           

        
            
    def merge_tokens(self, tokens):
        """
        Makes potential date strings out of matches, got from tokenize_string method.
        :param tokens: [(match_text, match_group, {match.capturesdict()}), ...]
        :return: potential date strings
        """
        MIN_MATCHES = 3
        fragments = []
        frag = DateFragment()
        start_index = 0
        last_delimiter = None
        current_is_delimiter = None
        next_delimiter = None
        num_tokens = len(tokens)
        start_char, total_chars = 0, 0
        stack_delim = self.stack_delimiters(tokens)
                                           
        for idx,token in enumerate(tokens):

            total_chars += len(token[0])
            # print("--------------------------<token>")
            
            
            tok_text, group, tok_capts = token[0], token[1], token[2]
            # print("token: ",tok_text, "grupo: ",group )
            
            # si no tiene grupo se reinicia frag 
            if not group:
                if frag.indices[1] > 0:
                    if frag.get_captures_count() >= MIN_MATCHES:
                        fragments.append(frag)
                frag = DateFragment()
                last_delimiter = 'no_group'
                start_index = idx +1
                start_char = total_chars
                # print("-----------------------<break no group>")
                continue
            
            
                
            # print("last delimiter: ", last_delimiter)

            if len(tok_capts['delimiters']) != 0:
                current_is_delimiter = True
                
                
                if idx + 1 <= num_tokens:
                    next_delimiter = self.find_next_delimiters(tokens[idx + 1:])
                    # print("next_delimiter: ", next_delimiter)

                    if next_delimiter is None :
                        next_delimiter = tok_capts['delimiters'][0]
                        # print("-----------------------<no hay delimiter >")
                        

                else: 
                    next_delimiter = tok_capts['delimiters'][0]
                    # print("-----------------------<delimiter $>")
                
                
                delimiters_list = self.find_delimiter(stack_delim, start_index,idx)
                # print("delimiters_list: ",delimiters_list)
                
                num_rep = delimiters_list.count(tok_capts['delimiters'][0])
                
                # print("num_rep:", num_rep)
                sentence_2 = (tok_capts['delimiters'][0] != next_delimiter )
                
                
                if sentence_2 and  num_rep != 2:
                    if frag.indices[1] > 0:
                        if frag.get_captures_count() >= MIN_MATCHES:
                            fragments.append(frag)
    
                    # print("-----------------------<break sentence and num_rep>")
                    frag = DateFragment()
                    last_delimiter = tok_capts['delimiters'][0]
                    start_index = idx +1
                    start_char = total_chars
                    continue
                
                last_delimiter = tok_capts['delimiters'][0]

            
            if frag.indices[1] == 0:
                frag.indices = (start_char, total_chars)
            else:
                frag.indices = (frag.indices[0], total_chars)  # -1

            frag.match_str += tok_text

            for capt in tok_capts:
                if capt in frag.captures:
                    frag.captures[capt] += tok_capts[capt]
                else:
                    frag.captures[capt] = tok_capts[capt]
                

            # print('frag actual: ',frag.match_str)
            start_char = total_chars
            
            # print("--------------------------------<token fin>\n\n")
                
            
            

        if frag.get_captures_count() >= MIN_MATCHES:  # frag.matches
            fragments.append(frag)

        for frag in fragments:
            for gr in ALL_GROUPS:
                if gr not in frag.captures:
                    frag.captures[gr] = []

        return fragments

    @staticmethod
    def get_token_group(captures):
        for gr in ALL_GROUPS:
            lst = captures.get(gr)
            if lst and len(lst) > 0:
                return gr
        return ""

    @staticmethod
    def split_date_range(text):
        st_matches = RANGE_SPLIT_REGEX.finditer(text)
        start = 0
        parts = []  # List[Tuple[str, Tuple[int, int]]]

        for match in st_matches:
            match_start = match.start()
            if match_start > start:
                parts.append((text[start:match_start], (start, match_start)))
            start = match.end()

        if start < len(text):
            parts.append((text[start:], (start, len(text))))

        return parts


def find_dates(text, source=True, index=True, strict=False, base_date=None, first="month"):
    """
    Extract datetime strings from text

    :param text:
        A string that contains one or more natural language or literal
        datetime strings
    :type text: str|unicode
    :param source:
        Return the original string segment
    :type source: boolean
    :param index:
        Return the indices where the datetime string was located in text
    :type index: boolean
    :param strict:
        Only return datetimes with complete date information. For example:
        `July 2016` of `Monday` will not return datetimes.
        `May 16, 2015` will return datetimes.
    :type strict: boolean
    :param base_date:
        Set a default base datetime when parsing incomplete dates
    :type base_date: datetime
    :param first:
        Whether to interpret the the first value in an ambiguous 3-integer date
        (01/02/03) as the month, day, or year. Values can be `month`, `day`, `year`.
        Default is `month`.
    :type first: str|unicode


    :return: Returns a generator that produces :mod:`datetime.datetime` objects,
        or a tuple with the source text and index, if requested
    """
    date_finder = DateFinder(base_date=base_date, first=first)
    return date_finder.find_dates(text, source=source, index=index, strict=strict)



import regex as re

NUMBERS_PATTERN = r"first|second|third|fourth|fifth|sixth|seventh|eighth|nineth|tenth"
POSITIONNAL_TOKENS = r"next_|last_"
DIGITS_PATTERN = r"\d+"
DIGITS_SUFFIXES = r"st|th|rd|nd"
DAYS_PATTERN = "monday|tuesday|wednesday|thursday|friday|saturday|sunday|mandag|tirsdag|onsdag|torsdag|fredag|lørdag|søndag|mon|tue|tues|wed|thu|thur|thurs|fri|sat|sun|man|tir|tirs|ons|tor|tors|fre|lør|søn"
MONTHS_PATTERN = r"january|february|march|april|may|june|july|august|september|october|november|december|enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|januar|februar|marts|april|maj|juni|juli|august|september|oktober|november|december|jan|ene|feb|marz|apr|abr|may|maj|jun|jul|aug|ago|sep|sept|oct|okt|nov|dec|dic"
TIMEZONES_PATTERN = "ACDT|ACST|ACT|ACWDT|ACWST|ADDT|ADMT|ADT|AEDT|AEST|AFT|AHDT|AHST|AKDT|AKST|AKTST|AKTT|ALMST|ALMT|AMST|AMT|ANAST|ANAT|ANT|APT|AQTST|AQTT|ARST|ART|ASHST|ASHT|AST|AWDT|AWST|AWT|AZOMT|AZOST|AZOT|AZST|AZT|BAKST|BAKT|BDST|BDT|BEAT|BEAUT|BIOT|BMT|BNT|BORT|BOST|BOT|BRST|BRT|BST|BTT|BURT|CANT|CAPT|CAST|CAT|CAWT|CCT|CDDT|CDT|CEDT|CEMT|CEST|CET|CGST|CGT|CHADT|CHAST|CHDT|CHOST|CHOT|CIST|CKHST|CKT|CLST|CLT|CMT|COST|COT|CPT|CST|CUT|CVST|CVT|CWT|CXT|ChST|DACT|DAVT|DDUT|DFT|DMT|DUSST|DUST|EASST|EAST|EAT|ECT|EDDT|EDT|EEDT|EEST|EET|EGST|EGT|EHDT|EMT|EPT|EST|ET|EWT|FET|FFMT|FJST|FJT|FKST|FKT|FMT|FNST|FNT|FORT|FRUST|FRUT|GALT|GAMT|GBGT|GEST|GET|GFT|GHST|GILT|GIT|GMT|GST|GYT|HAA|HAC|HADT|HAE|HAP|HAR|HAST|HAT|HAY|HDT|HKST|HKT|HLV|HMT|HNA|HNC|HNE|HNP|HNR|HNT|HNY|HOVST|HOVT|HST|ICT|IDDT|IDT|IHST|IMT|IOT|IRDT|IRKST|IRKT|IRST|ISST|IST|JAVT|JCST|JDT|JMT|JST|JWST|KART|KDT|KGST|KGT|KIZST|KIZT|KMT|KOST|KRAST|KRAT|KST|KUYST|KUYT|KWAT|LHDT|LHST|LINT|LKT|LMT|LMT|LMT|LMT|LRT|LST|MADMT|MADST|MADT|MAGST|MAGT|MALST|MALT|MART|MAWT|MDDT|MDST|MDT|MEST|MET|MHT|MIST|MIT|MMT|MOST|MOT|MPT|MSD|MSK|MSM|MST|MUST|MUT|MVT|MWT|MYT|NCST|NCT|NDDT|NDT|NEGT|NEST|NET|NFT|NMT|NOVST|NOVT|NPT|NRT|NST|NT|NUT|NWT|NZDT|NZMT|NZST|OMSST|OMST|ORAST|ORAT|PDDT|PDT|PEST|PET|PETST|PETT|PGT|PHOT|PHST|PHT|PKST|PKT|PLMT|PMDT|PMMT|PMST|PMT|PNT|PONT|PPMT|PPT|PST|PT|PWT|PYST|PYT|QMT|QYZST|QYZT|RET|RMT|ROTT|SAKST|SAKT|SAMT|SAST|SBT|SCT|SDMT|SDT|SET|SGT|SHEST|SHET|SJMT|SLT|SMT|SRET|SRT|SST|STAT|SVEST|SVET|SWAT|SYOT|TAHT|TASST|TAST|TBIST|TBIT|TBMT|TFT|THA|TJT|TKT|TLT|TMT|TOST|TOT|TRST|TRT|TSAT|TVT|ULAST|ULAT|URAST|URAT|UTC|UYHST|UYST|UYT|UZST|UZT|VET|VLAST|VLAT|VOLST|VOLT|VOST|VUST|VUT|WARST|WART|WAST|WAT|WDT|WEDT|WEMT|WEST|WET|WFT|WGST|WGT|WIB|WIT|WITA|WMT|WSDT|WSST|WST|WT|XJT|YAKST|YAKT|YAPT|YDDT|YDT|YEKST|YEKST|YEKT|YEKT|YERST|YERT|YPT|YST|YWT|zzz"
## explicit north american timezones that get replaced
NA_TIMEZONES_PATTERN = "pacific|eastern|mountain|central"
ALL_TIMEZONES_PATTERN = TIMEZONES_PATTERN + "|" + NA_TIMEZONES_PATTERN

# Allows for straightforward datestamps e.g 2017, 201712, 20171223. Created with:
#  YYYYMM_PATTERN = '|'.join(['19\d\d'+'{:0>2}'.format(mon)+'|20\d\d'+'{:0>2}'.format(mon) for mon in range(1, 13)])
#  YYYYMMDD_PATTERN = '|'.join(['19\d\d'+'{:0>2}'.format(mon)+'[0123]\d|20\d\d'+'{:0>2}'.format(mon)+'[0123]\d' for mon in range(1, 13)])

DELIMITERS_PATTERN = r"[/\:\-\,\.\s\_]+"

YYYY_PATTERN = r"202\d|202|20"
DD_PATTERN = r"[0123]\d"




YYYYMM_PATTERN = r"19\d\d01|20\d\d01|19\d\d02|20\d\d02|19\d\d03|20\d\d03|19\d\d04|20\d\d04|19\d\d05|20\d\d05|19\d\d06|20\d\d06|19\d\d07|20\d\d07|19\d\d08|20\d\d08|19\d\d09|20\d\d09|19\d\d10|20\d\d10|19\d\d11|20\d\d11|19\d\d12|20\d\d12"

YYYYMMDD_PATTERN = r"19\d\d01[0123]\d|20\d\d01[0123]\d|19\d\d02[0123]\d|20\d\d02[0123]\d|19\d\d03[0123]\d|20\d\d03[0123]\d|19\d\d04[0123]\d|20\d\d04[0123]\d|19\d\d05[0123]\d|20\d\d05[0123]\d|19\d\d06[0123]\d|20\d\d06[0123]\d|19\d\d07[0123]\d|20\d\d07[0123]\d|19\d\d08[0123]\d|20\d\d08[0123]\d|19\d\d09[0123]\d|20\d\d09[0123]\d|19\d\d10[0123]\d|20\d\d10[0123]\d|19\d\d11[0123]\d|20\d\d11[0123]\d|19\d\d12[0123]\d|20\d\d12[0123]\d"

YYYYMMDDHHMMSS_PATTERN = "|".join(
    [
        r"19\d\d"
        + "{:0>2}".format(mon)
        + r"[0-3]\d[0-5]\d[0-5]\d[0-5]\d|20\d\d"
        + "{:0>2}".format(mon)
        + r"[0-3]\d[0-5]\d[0-5]\d[0-5]\d"
        for mon in range(1, 13)
    ]
)
ISO8601_PATTERN = r"(?P<years>-?(\:[1-9][0-9]*)?[0-9]{4})\-(?P<months>1[0-2]|0[1-9])\-(?P<days>3[01]|0[1-9]|[12][0-9])T(?P<hours>2[0-3]|[01][0-9])\:(?P<minutes>[0-5][0-9]):(?P<seconds>[0-5][0-9])(?:[\.,]+(?P<microseconds>[0-9]+))?(?P<offset>(?:Z|[+-](?:2[0-3]|[01][0-9])\:[0-5][0-9]))?"

UNDELIMITED_STAMPS_PATTERN = "|".join([YYYYMMDDHHMMSS_PATTERN, YYYYMMDD_PATTERN, YYYYMM_PATTERN, ISO8601_PATTERN])

DELIMITERS_PATTERN = r"[/\:\-\,\.\s\_]+"

TIME_PERIOD_PATTERN = r"a\.m\.|am|p\.m\.|pm"

## can be in date strings but not recognized by dateutils
EXTRA_TOKENS_PATTERN = r"\sal\s|during|standard|daylight|savings|time|date|dated|through|between|until|day"

## TODO: Get english numbers?
## http://www.rexegg.com/regex-trick-numbers-in-english.html

RELATIVE_PATTERN = "before|after_|next_|last_"
TIME_SHORTHAND_PATTERN = "noon_|midnight_|today_|yesterday_"
UNIT_PATTERN = "second_|minute_|hour_|month_|year_"

## Time pattern is used independently, so specified here.
TIME_PATTERN = r"""
(?P<time>
    ## Captures in format XX:YY(:ZZ) (PM) (EST)
    (
        (?P<hours>\d{{1,2}})
        \:
        (?P<minutes>\d{{1,2}})
        (\:(?<seconds>\d{{1,2}}))?
        ([\.\,](?<microseconds>\d{{1,6}}))?
        \s*
        (?P<time_periods>{time_periods})?
        \s*
        (?P<timezones>{timezones})?
    )
    |
    ## Captures in format 11 AM (EST)
    ## Note with single digit capture requires time period
    (
        (?P<hours>\d{{1,2}})
        \s*
        (?P<time_periods>{time_periods})
        \s*
        (?P<timezones>{timezones})*
    )
)
""".format(
    time_periods=TIME_PERIOD_PATTERN, timezones=ALL_TIMEZONES_PATTERN
)

DATES_PATTERN = """
((?P<undelimited_stamps>{undelimited_stamps})|({time}|
        ## Grab any four digit years
        (?P<years>{years})
        |
        ## Numbers
        (?P<numbers>{numbers})
        ## Grab any digits
        |
        (?P<digits>{digits})(?P<digits_suffixes>{digits_suffixes})?
        |
        (?P<days>{days})
        |
        (?P<months>{months})
        |
        ## Delimiters, ie Tuesday[,] July 18 or 6[/]17[/]2008
        ## as well as whitespace
        (?P<delimiters>{delimiters})
        |
        (?P<positionnal_tokens>{positionnal_tokens})
        |
        ## These tokens could be in phrases that dateutil does not yet recognize
        ## Some are US Centric
        (?P<extra_tokens>{extra_tokens})
    ## We need at least three items to match for minimal datetime parsing
    ## ie 10pm
    ){{1,1}}
)
"""

DATES_PATTERN = DATES_PATTERN.format(
    time=TIME_PATTERN,
    undelimited_stamps=UNDELIMITED_STAMPS_PATTERN,
    years=YYYY_PATTERN,
    numbers=NUMBERS_PATTERN,
    digits=DIGITS_PATTERN,
    digits_suffixes=DIGITS_SUFFIXES,
    days=DAYS_PATTERN,
    months=MONTHS_PATTERN,
    delimiters=DELIMITERS_PATTERN,
    positionnal_tokens=POSITIONNAL_TOKENS,
    extra_tokens=EXTRA_TOKENS_PATTERN,
)

ALL_GROUPS = ['time', 'years', 'numbers', 'digits', 'digits_suffixes', 'days',
              'months', 'delimiters', 'positionnal_tokens', 'extra_tokens',
              'undelimited_stamps', 'hours', 'minutes', 'seconds', 'microseconds',
              'time_periods', 'timezones']

DATE_REGEX = re.compile(
    DATES_PATTERN, re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.VERBOSE
)

TIME_REGEX = re.compile(
    TIME_PATTERN, re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.VERBOSE
)

## These tokens can be in original text but dateutil
## won't handle them without modification
REPLACEMENTS = {
    "standard": " ",
    "daylight": " ",
    "savings": " ",
    "time": " ",
    "date": " ",
    "by_": " ",
    "due_": " ",
    "on_": " ",
    "to_": " ",
    "day": " ",
}

TIMEZONE_REPLACEMENTS = {
    "pacific": "PST",
    "eastern": "EST",
    "mountain": "MST",
    "central": "CST",
}

## Characters that can be removed from ends of matched strings
STRIP_CHARS = " \n\t:-.,_"

# split ranges
RANGE_SPLIT_PATTERN = r'\Wal\W'

RANGE_SPLIT_REGEX =  re.compile(RANGE_SPLIT_PATTERN,
    re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL)





from pymongo import MongoClient
from random import randint
# from Crypto.Cipher import AES
# from Crypto.Util.Padding import pad, unpad
from sys import stdout
from dateutil import parser
from psycopg import connect
from os import getenv
from klp_commons.log.structuredMessage import StructuredMessage
from logging import getLogger
from dateutil.parser import parse
from pandas import DataFrame
import traceback
from pickle import load as load_
# from report.cryptology import Cryptology
from pymongo.errors import DuplicateKeyError, BulkWriteError
from json import loads
from os import remove
fmt = StructuredMessage
message = 'mongoDBController'

class ControllerMongoDB:
    """
    for timestamp use :
    
        expiry_date = '2021-07-13T00:00:00.000Z'
        expiry = parser.parse(expiry_date)

    """
    # Variables de clase

    def __init__(self, pad: int = None):
        self.logger = getLogger("ControllerMongo")

        self.MONGODB_URL = getenv("MONGODB_URL")        
        self.data = None
        self.client = None
        self.con_db = None
        self.format_exter = '%Y-%m-%dT%H:%M:%S.%f%z'
        self.format_inter = '%Y-%m-%dT%H:%M:%S'
        self.logger.info("done creating an instance of ControllerMongo")

    def get_con_mongo(self):
        try:
            self.logger.info("createnig connection MongoClient")

            self.client = MongoClient(self.MONGODB_URL,authSource="admin")

        except Exception as e:
            self.logger.error("Exception createnig connection MongoClient seen: " + str(e))
            traceback.print_exc(file = stdout)
            
        self.logger.info("done createnig connection MongoClient")

    def close_con(self):
        try:
            self.logger.info("close_con()  ControllerMongo")

            self.client.close()
        except Exception as e:
            self.logger.error("Exception close_con() ControllerMongo seen: " + str(e))
            traceback.print_exc(file = stdout)
        
        self.logger.info("done close_con()  ControllerMongo")
         
    def get_list_dbs(self):
        return self.client.list_database_names()

    def get_con_db(self,name_db):
        self.get_con_mongo()
        self.con_db  = self.client[name_db]

    def get_con_collect(self,collection_name,db_name="klopp"):
        self.get_con_mongo()
        self.con_db  = self.client[db_name][collection_name]

    def insert_one_fact(self):
        try:
            self.con_db.fact.insert_one(self.data)
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.client.close()
            
    def exist(self,key,value):
        try:
            return bool(self.con_db.find_one({key:value}))
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.client.close()
            
    def insert_one_collection(self):
        try:
            self.con_db.insert_one(self.data)
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.client.close()

    def insert_many_fact(self):
        
        try:
            self.con_db.fact.insert_many(self.data)

        except (DuplicateKeyError, BulkWriteError) as e:

            self.logger.exception(e, exc_info=True)
            self.logger.info("Save into error colecction mongodb ")
            self.insert_error()
            self.logger.info("Done save into error colecction mongodb ")

        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.client.close()

    def insert_error(self):
        try:
            self.con_db.errorDuplicateKey.insert_one({"error":self.data})   
        except Exception as e:
            self.logger.exception(e, exc_info=True)

    def insert_error_ms_report(self):
        try:
            self.con_db.errorMSReport.insert_one({"error":self.data})   
        except Exception as e:
            self.logger.exception(e, exc_info=True)

    def insert_error_ms_cat(self):
        try:
            self.con_db.errorMSCat.insert_one({"error":self.data})   
        except Exception as e:
            self.logger.exception(e, exc_info=True)

    def insert_errorNoFound(self):

        try:

            self.con_db.errorSubcategoryNoFound.insert_one({"error":self.data})   
        except Exception as e:
            self.logger.error(e, exc_info=True)

    def insert_many_analytic(self):
        try:
            self.con_db.analytic.insert_many(self.data)

        except (DuplicateKeyError, BulkWriteError) as e:
            self.logger.error(e, exc_info=True)
            self.logger.info("Save into error colecction mongodb ")
            self.insert_error()
            self.logger.info("Done save into error colecction mongodb ")

        except Exception as e:
            self.logger.error(e, exc_info=True)
        finally:
            self.client.close()

    def update_one_fact(self,ref, new_item):
        try:
            query = { "_id": ref}
            new_item = {'$push': {'history-track': new_item}}
            self.con_db.fact.update_one(query,new_item)
        
        except Exception as e:
            self.logger.error(e, exc_info=True)
        finally:
            self.client.close()

    def set_document(self,data):
        self.data = data  

    def ingesta_mongo(self,key,pad,fact_path,path_analytic):
        import datetime
        from utils.utils import divide_chunks

        crypto = Cryptology(pad)

        crypto.set_key(key)


        # Load data (deserialize)
        with open(fact_path, 'rb') as handle:
            data = load_(handle)

        fact = crypto.decrypt(data)

        self.data = loads(fact)

        for item in self.data:
            item['created_at'] = datetime.strptime(item['created_at'].split('+')[0].split('.')[0], self.format_extern).replace(microsecond=0).isoformat()


        docs_array = divide_chunks(list(self.data))
        for item in docs_array:
            self.data = item
            self.get_con_db('klopp')
            self.insert_many_fact()

        # Load data (deserialize)
        with open(path_analytic, 'rb') as handle:
            data = load_(handle)

        analytic = crypto.decrypt(data)
        self.data = loads(analytic)

        docs_array = divide_chunks(list(self.data))
        for item in docs_array:
            self.data = item
            self.get_con_db('klopp')
            self.insert_many_analytic()


        


        remove(fact_path)
        remove(path_analytic)


from kafka import KafkaProducer
from json import dumps as dumps_


class Producer:
    def __init__(
                self, 
                topic = None,
                kafka_brokers = None,
                acks = None,
                api_version = None,
                client_id = None
                 ):
        self.kafka_api_version = api_version
        self.topic = topic
        self.acks = acks
        self.client_id = client_id
        self.kafka_brokers = kafka_brokers
        
    def init_producer(self) -> None:

        self.producer = KafkaProducer(
            acks = self.acks,
            api_version=self.kafka_api_version,
            bootstrap_servers = self.kafka_brokers,
            value_serializer = lambda x: dumps_(x).encode('utf-8'),
        )

    def send_message(self, data: dict) -> None:
        self.producer.send(self.topic, value=data)
        # block until all async messages are sent
        self.producer.flush()
